"""Migration the old friction_type 4 to 2 (MANNING)

Revision ID: 0201
Revises:
Create Date: 2021-09-29 13:50:19.544275

"""
from alembic import op

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "0201"
down_revision = "0200"
branch_labels = None
depends_on = None

TABLES = ("v2_cross_section_location", "v2_pipe", "v2_culvert", "v2_weir", "v2_orifice")
COLUMN_NAME = "friction_type"


def upgrade():
    for table_name in TABLES:
        upgrade_single_table(table_name)


def downgrade():
    pass


def upgrade_single_table(table_name):
    table = sa.table(
        table_name,
        sa.column("friction_type", sa.Integer),
    )
    op.execute(
        table.update()
        .where(table.c.friction_type == op.inline_literal(4))
        .values({"friction_type": op.inline_literal(2)})
    )



"""global raster values

Revision ID: 0210
Revises: 0209
Create Date: 2022-11-08 16:37

"""
from alembic import op

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "0210"
down_revision = "0209"
branch_labels = None
depends_on = None


MIGRATION_QUERIES = """
UPDATE v2_interflow SET hydraulic_conductivity = NULL WHERE hydraulic_conductivity < 0;
UPDATE v2_groundwater SET initial_infiltration_rate = NULL WHERE initial_infiltration_rate < 0;
UPDATE v2_groundwater SET equilibrium_infiltration_rate = NULL WHERE equilibrium_infiltration_rate < 0;
UPDATE v2_groundwater SET infiltration_decay_period = NULL WHERE infiltration_decay_period <= 0;
UPDATE v2_groundwater SET groundwater_hydro_connectivity = NULL WHERE groundwater_hydro_connectivity < 0;
"""


def upgrade():
    for q in MIGRATION_QUERIES.split(";"):
        op.execute(q)

    with op.batch_alter_table("v2_simple_infiltration") as batch_op:
        batch_op.add_column(
            sa.Column("max_infiltration_capacity", sa.Float(), nullable=True)
        )


def downgrade():
    with op.batch_alter_table("v2_simple_infiltration") as batch_op:
        batch_op.drop_column("max_infiltration_capacity")



"""add breach and exchange properties

Revision ID: 0212
Revises: 0211
Create Date: 2022-12-12 14:48:00

"""
from alembic import op

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "0212"
down_revision = "0211"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("v2_exchange_line") as batch_op:
        batch_op.add_column(sa.Column("exchange_level", sa.Float(), nullable=True))
    with op.batch_alter_table("v2_potential_breach") as batch_op:
        batch_op.add_column(sa.Column("exchange_level", sa.Float(), nullable=True))
        batch_op.add_column(sa.Column("levee_material", sa.Integer(), nullable=True))
        batch_op.add_column(
            sa.Column("maximum_breach_depth", sa.Float(), nullable=True)
        )


def downgrade():
    with op.batch_alter_table("v2_exchange_line") as batch_op:
        batch_op.drop_column("exchange_level")
    with op.batch_alter_table("v2_potential_breach") as batch_op:
        batch_op.drop_column("exchange_level")
        batch_op.drop_column("levee_material")
        batch_op.drop_column("maximum_breach_depth")



from enum import Enum
from sqlalchemy.orm import Query
from sqlalchemy.orm.session import Session
from threedi_modelchecker.simulation_templates.exceptions import (
    SchematisationError,
)
from threedi_modelchecker.simulation_templates.utils import parse_timeseries
from threedi_modelchecker.threedi_model.models import BoundaryCondition1D
from threedi_modelchecker.threedi_model.models import BoundaryConditions2D
from typing import Dict
from typing import List
from typing import Union


# JSON format example:
# [
#     {
#         "id": 1,  (spatialite id from boundary table)
#         "type": "2D"
#         "interpolate": true,
#         "values": [
#             [0, 0.5],
#             [500, 0,8],
#             [1000, 0]
#         ]
#     },
#     {
#         "id": 2, (spatialite id from boundary table)
#         "type": "1D",
#         "interpolate": true,
#         "values": [
#             [0, 0,3],
#             [400, 0.1]
#         ]
#     },
#     {
#         "id": 3, (spatialite id from boundary table)
#         "type": "1D",
#         "interpolate": true,
#         "values": [
#             [0, -2.4],
#             [1300, 0,3],
#             [3000, 1.2],
#             [3600, 0]
#         ]
#     }
# ]


class BoundaryType(Enum):
    one_d = "1D"
    two_d = "2D"


def sqlite_boundary_to_dict(
    boundary: Union[BoundaryConditions2D, BoundaryCondition1D]
) -> Dict:

    boundary_1d2d: BoundaryType = BoundaryType.one_d
    if isinstance(boundary, BoundaryConditions2D):
        boundary_1d2d = BoundaryType.two_d

    try:
        values = parse_timeseries(boundary.timeseries)
    except (TypeError, ValueError):
        raise SchematisationError(
            f"Incorrect formatted timeseries for {boundary_1d2d.value} boundary condition with id={boundary.id}"
        )

    return {
        "id": boundary.id,
        "type": boundary_1d2d.value,
        "interpolate": True,
        "values": values,
    }


class BoundariesExtractor(object):
    def __init__(self, session: Session):
        self.session = session
        self._boundaries_2d = None
        self._boundaries_1d = None

    @property
    def boundaries_2d(self) -> List[Dict]:
        if self._boundaries_2d is None:
            boundaries_2d = (
                Query(BoundaryConditions2D)
                .with_session(self.session)
                .order_by(BoundaryConditions2D.id)
                .all()
            )

            self._boundaries_2d = [sqlite_boundary_to_dict(x) for x in boundaries_2d]

        return self._boundaries_2d

    @property
    def boundaries_1d(self) -> List[Dict]:
        if self._boundaries_1d is None:
            boundaries_1d = (
                Query(BoundaryCondition1D)
                .with_session(self.session)
                .order_by(BoundaryCondition1D.id)
                .all()
            )
            self._boundaries_1d = [sqlite_boundary_to_dict(x) for x in boundaries_1d]

        return self._boundaries_1d

    def as_list(self) -> List[Dict]:
        """
        Returns: list with dict's for every boundary, 2d boundaries before 1d boundaries
        """
        return self.boundaries_2d + self.boundaries_1d



"""breach and exchange

Revision ID: 0211
Revises: 0210
Create Date: 2022-11-23 11:21:10.967235

"""
from alembic import op

import geoalchemy2
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "0211"
down_revision = "0210"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "v2_exchange_line",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("channel_id", sa.Integer(), nullable=True),
        sa.Column(
            "the_geom",
            geoalchemy2.types.Geometry(
                geometry_type="LINESTRING",
                srid=4326,
                management=True,
                from_text="ST_GeomFromEWKT",
                name="geometry",
            ),
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "v2_potential_breach",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("code", sa.String(length=100), nullable=True),
        sa.Column("display_name", sa.String(length=255), nullable=True),
        sa.Column("channel_id", sa.Integer(), nullable=True),
        sa.Column(
            "the_geom",
            geoalchemy2.types.Geometry(
                geometry_type="LINESTRING",
                srid=4326,
                management=True,
                from_text="ST_GeomFromEWKT",
                name="geometry",
            ),
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("id"),
    )


def downgrade():
    op.drop_table("v2_potential_breach")
    op.drop_table("v2_exchange_line")



from ..exceptions import SchematisationError
from sqlalchemy.orm import Query
from sqlalchemy.orm.session import Session
from threedi_api_client.openapi.models import GroundWaterRaster
from threedi_api_client.openapi.models import OneDWaterLevel
from threedi_api_client.openapi.models import TwoDWaterRaster
from threedi_api_client.openapi.models.ground_water_level import (
    GroundWaterLevel,
)
from threedi_api_client.openapi.models.one_d_water_level_file import (
    OneDWaterLevelFile,
)
from threedi_api_client.openapi.models.two_d_water_level import TwoDWaterLevel
from threedi_modelchecker.simulation_templates.models import InitialWaterlevels
from threedi_modelchecker.threedi_model.models import ConnectionNode
from threedi_modelchecker.threedi_model.models import GlobalSetting
from typing import Optional


sqlite_agg_method_to_api_map = {0: "max", 1: "min", 2: "mean"}


class InitialWaterlevelExtractor(object):
    def __init__(self, session: Session, global_settings_id: Optional[int] = None):
        self.session = session
        self._global_settings = None
        self._global_settings_id = global_settings_id
        self._connection_nodes_with_initial_waterlevels = None

    @property
    def has_connection_nodes_with_initial_waterlevels(self) -> bool:
        if self._connection_nodes_with_initial_waterlevels is None:
            self._connection_nodes_with_initial_waterlevels = (
                Query(ConnectionNode)
                .with_session(self.session)
                .filter(ConnectionNode.initial_waterlevel != None)
                .first()
                is not None
            )
        return self._connection_nodes_with_initial_waterlevels

    @property
    def global_settings(self) -> GlobalSetting:
        if self._global_settings is None:
            qr = Query(GlobalSetting).with_session(self.session)
            if self._global_settings_id is not None:
                qr = qr.filter(GlobalSetting.id == self._global_settings_id)
            self._global_settings = qr.first()
        return self._global_settings

    @property
    def constant_waterlevel_1d(self) -> Optional[OneDWaterLevel]:
        if self.global_settings.initial_waterlevel is None:
            return None
        if self.has_connection_nodes_with_initial_waterlevels:
            return None
        if float(self.global_settings.initial_waterlevel) == -9999:
            return None
        return OneDWaterLevel(value=float(self.global_settings.initial_waterlevel))

    @property
    def constant_waterlevel_2d(self) -> Optional[TwoDWaterLevel]:
        if self.global_settings.initial_waterlevel is None:
            return None
        if self.global_settings.initial_waterlevel_file:
            return None
        if float(self.global_settings.initial_waterlevel) == -9999:
            return None
        return TwoDWaterLevel(value=float(self.global_settings.initial_waterlevel))

    @property
    def constant_waterlevel_groundwater(self) -> Optional[GroundWaterLevel]:
        if self.global_settings.initial_groundwater_level is None:
            return None
        if self.global_settings.initial_groundwater_level_file:
            return None
        if float(self.global_settings.initial_groundwater_level) == -9999:
            return None
        return GroundWaterLevel(
            value=float(self.global_settings.initial_groundwater_level)
        )

    @property
    def waterlevel_1d_file(self) -> Optional[OneDWaterLevelFile]:
        if not self.has_connection_nodes_with_initial_waterlevels:
            return None
        return OneDWaterLevelFile(initial_waterlevel="resolved_during_saving")

    @property
    def waterlevel_2d_raster(self) -> Optional[TwoDWaterRaster]:
        if not self.global_settings.initial_waterlevel_file:
            return None
        if self.global_settings.water_level_ini_type is None:
            raise SchematisationError(
                "initial_waterlevel_file supplied without water_level_ini_type"
            )
        return TwoDWaterRaster(
            aggregation_method=sqlite_agg_method_to_api_map[
                self.global_settings.water_level_ini_type.value
            ],
            initial_waterlevel=self.global_settings.initial_waterlevel_file,
        )

    @property
    def waterlevel_groundwater_raster(self) -> Optional[GroundWaterRaster]:
        if not self.global_settings.initial_groundwater_level_file:
            return None
        if self.global_settings.initial_groundwater_level_type is None:
            raise SchematisationError(
                "initial_groundwater_level_file supplied without initial_groundwater_level_type"
            )
        return GroundWaterRaster(
            aggregation_method=sqlite_agg_method_to_api_map[
                self.global_settings.initial_groundwater_level_type.value
            ],
            initial_waterlevel=self.global_settings.initial_groundwater_level_file,
        )

    def all_initial_waterlevels(self) -> InitialWaterlevels:
        return InitialWaterlevels(
            self.constant_waterlevel_2d,
            self.constant_waterlevel_1d,
            self.constant_waterlevel_groundwater,
            self.waterlevel_1d_file,
            self.waterlevel_2d_raster,
            self.waterlevel_groundwater_raster,
        )



from ..threedi_model import custom_types
from .base import EnumCheck
from .base import ForeignKeyCheck
from .base import GeometryCheck
from .base import GeometryTypeCheck
from .base import NotNullCheck
from .base import TypeCheck
from .base import UniqueCheck
from geoalchemy2.types import Geometry


def get_level(table, column, level_map):
    level = level_map.get(f"*.{column.name}")
    level = level_map.get(f"{table.name}.{column.name}", level)
    return level or "ERROR"


def generate_foreign_key_checks(table, custom_level_map=None, **kwargs):
    custom_level_map = custom_level_map or {}
    foreign_key_checks = []
    for fk_column in table.foreign_keys:
        level = get_level(table, fk_column.parent, custom_level_map)
        foreign_key_checks.append(
            ForeignKeyCheck(
                reference_column=fk_column.column,
                column=fk_column.parent,
                level=level,
                **kwargs,
            )
        )
    return foreign_key_checks


def generate_unique_checks(table, custom_level_map=None, **kwargs):
    custom_level_map = custom_level_map or {}
    unique_checks = []
    for column in table.columns:
        if column.unique or column.primary_key:
            level = get_level(table, column, custom_level_map)
            unique_checks.append(UniqueCheck(column, level=level, **kwargs))
    return unique_checks


def generate_not_null_checks(table, custom_level_map=None, **kwargs):
    custom_level_map = custom_level_map or {}
    not_null_checks = []
    for column in table.columns:
        if not column.nullable:
            level = get_level(table, column, custom_level_map)
            not_null_checks.append(NotNullCheck(column, level=level, **kwargs))
    return not_null_checks


def generate_type_checks(table, custom_level_map=None, **kwargs):
    custom_level_map = custom_level_map or {}
    data_type_checks = []
    for column in table.columns:
        level = get_level(table, column, custom_level_map)
        data_type_checks.append(TypeCheck(column, level=level, **kwargs))
    return data_type_checks


def generate_geometry_checks(table, custom_level_map=None, **kwargs):
    custom_level_map = custom_level_map or {}
    geometry_checks = []
    for column in table.columns:
        if type(column.type) == Geometry:
            level = get_level(table, column, custom_level_map)
            geometry_checks.append(GeometryCheck(column, level=level, **kwargs))
    return geometry_checks


def generate_geometry_type_checks(table, custom_level_map=None, **kwargs):
    custom_level_map = custom_level_map or {}
    geometry_type_checks = []
    for column in table.columns:
        if type(column.type) == Geometry:
            level = get_level(table, column, custom_level_map)
            geometry_type_checks.append(
                GeometryTypeCheck(column, level=level, **kwargs)
            )
    return geometry_type_checks


def generate_enum_checks(table, custom_level_map=None, **kwargs):
    custom_level_map = custom_level_map or {}
    enum_checks = []
    for column in table.columns:
        if issubclass(type(column.type), custom_types.CustomEnum):
            level = get_level(table, column, custom_level_map)
            enum_checks.append(EnumCheck(column, level=level, **kwargs))
    return enum_checks



from alembic import context
from threedi_modelchecker import ThreediDatabase
from threedi_modelchecker.threedi_model import constants
from threedi_modelchecker.threedi_model.models import Base

import os
import threedi_modelchecker.threedi_model.models  # NOQA needed for autogenerate


target_metadata = Base.metadata
config = context.config


def get_url():
    db_url = os.environ.get("DB_URL")
    if not db_url:
        raise RuntimeError(
            "Database URL must be specified using the environment variable DB_URL"
        )
    return db_url


def run_migrations_online():
    """Run migrations in 'online' mode.

    Note: SQLite does not (completely) support transactions, so, backup the
    SQLite before running migrations.
    """
    connectable = config.attributes.get("connection")
    if connectable is None:
        connectable = ThreediDatabase(get_url()).get_engine()

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            version_table=constants.VERSION_TABLE_NAME,
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    raise ValueError("Offline mode is not supported")
else:
    run_migrations_online()



from enum import Enum
from sqlalchemy.orm import Query
from sqlalchemy.orm.session import Session
from threedi_api_client.openapi.models import AggregationSettings
from threedi_api_client.openapi.models import NumericalSettings
from threedi_api_client.openapi.models import TimeStepSettings
from threedi_api_client.openapi.models.physical_settings import (
    PhysicalSettings,
)
from threedi_modelchecker.simulation_templates.models import Settings
from threedi_modelchecker.simulation_templates.settings.mappings import (
    aggregation_settings_map,
)
from threedi_modelchecker.simulation_templates.settings.mappings import (
    numerical_settings_map,
)
from threedi_modelchecker.simulation_templates.settings.mappings import (
    physical_settings_map,
)
from threedi_modelchecker.simulation_templates.settings.mappings import (
    sqlalchemy_to_api_model,
)
from threedi_modelchecker.simulation_templates.settings.mappings import (
    time_step_settings_map,
)
from threedi_modelchecker.threedi_model.constants import FlowVariable
from threedi_modelchecker.threedi_model.models import (
    AggregationSettings as SQLAggregationSettings,
)
from threedi_modelchecker.threedi_model.models import GlobalSetting
from threedi_modelchecker.threedi_model.models import (
    NumericalSettings as SQLNumericalSettings,
)
from typing import List
from typing import Optional


class SettingsExtractor(object):
    def __init__(self, session: Session, global_settings_id: Optional[int] = None):
        self.session = session
        self._global_settings = None
        self._global_settings_id = global_settings_id
        self._numerical_settings = None
        self._aggregation_settings = None

    @property
    def _sql_aggregation_settings(self) -> List[SQLAggregationSettings]:
        if self._aggregation_settings is None:
            self._aggregation_settings = (
                Query(SQLAggregationSettings)
                .with_session(self.session)
                .filter(
                    SQLAggregationSettings.global_settings_id == self.global_settings.id
                )
                .all()
            )
            if self._aggregation_settings is None:
                self._aggregation_settings = []
        return self._aggregation_settings

    @property
    def _sql_numerical_settings(self) -> SQLNumericalSettings:
        if self._numerical_settings is None:
            self._numerical_settings = (
                Query(SQLNumericalSettings)
                .with_session(self.session)
                .filter(
                    SQLNumericalSettings.id
                    == self.global_settings.numerical_settings_id
                )
                .first()
            )
        return self._numerical_settings

    @property
    def global_settings(self) -> GlobalSetting:
        if self._global_settings is None:
            qr = Query(GlobalSetting).with_session(self.session)
            if self._global_settings_id is not None:
                qr = qr.filter(GlobalSetting.id == self._global_settings_id)
            self._global_settings = qr.first()
        return self._global_settings

    @property
    def timestep_settings(self) -> TimeStepSettings:
        config_dict = {
            x.name: getattr(self.global_settings, x.name)
            for x in self.global_settings.__table__.columns
        }
        return sqlalchemy_to_api_model(
            config_dict, TimeStepSettings, time_step_settings_map
        )

    @property
    def physical_settings(self) -> PhysicalSettings:
        config_dict = {
            x.name: getattr(self.global_settings, x.name)
            for x in self.global_settings.__table__.columns
        }
        return sqlalchemy_to_api_model(
            config_dict, PhysicalSettings, physical_settings_map
        )

    @property
    def aggregation_settings(self) -> List[AggregationSettings]:
        agg_settings = []

        def convert_enum(value):
            if isinstance(value, FlowVariable):
                # Mismatch between sqlite and API....
                mapping = {
                    "waterlevel": "water_level",
                    "wet_cross-section": "wet_cross_section",
                }
                value = value.value
                value = mapping.get(value, value)
            elif isinstance(value, Enum):
                value = value.value
            return value

        for agg_setting in self._sql_aggregation_settings:
            agg_setting: SQLAggregationSettings
            config_dict = {
                x.name: convert_enum(getattr(agg_setting, x.name))
                for x in agg_setting.__table__.columns
            }
            agg_settings.append(
                sqlalchemy_to_api_model(
                    config_dict,
                    AggregationSettings,
                    aggregation_settings_map,
                    extra_exclude={
                        "name",
                    },
                )
            )
        return agg_settings

    @property
    def numerical_settings(self) -> NumericalSettings:
        config_dict = {
            x.name: getattr(self.global_settings, x.name)
            for x in self.global_settings.__table__.columns
        }
        config_dict.update(
            {
                x.name: getattr(self._sql_numerical_settings, x.name)
                for x in self._sql_numerical_settings.__table__.columns
            }
        )
        return sqlalchemy_to_api_model(
            config_dict, NumericalSettings, numerical_settings_map
        )

    def all_settings(self) -> Settings:
        return Settings(
            numerical=self.numerical_settings,
            physical=self.physical_settings,
            timestep=self.timestep_settings,
            aggregations=self.aggregation_settings,
        )



from datetime import timedelta
from sqlalchemy.orm import Query
from sqlalchemy.orm.session import Session
from threedi_api_client.openapi.models.measure_location import MeasureLocation
from threedi_api_client.openapi.models.measure_specification import (
    MeasureSpecification,
)
from threedi_api_client.openapi.models.memory_structure_control import (
    MemoryStructureControl,
)
from threedi_api_client.openapi.models.table_structure_control import (
    TableStructureControl,
)
from threedi_modelchecker.simulation_templates.exceptions import (
    SchematisationError,
)
from threedi_modelchecker.simulation_templates.models import StructureControls
from threedi_modelchecker.simulation_templates.utils import (
    strip_dict_none_values,
)
from threedi_modelchecker.threedi_model.constants import (
    ControlTableActionTypes,
)
from threedi_modelchecker.threedi_model.constants import ControlType
from threedi_modelchecker.threedi_model.constants import (
    MeasureLocationContentTypes,
)
from threedi_modelchecker.threedi_model.constants import MeasureVariables
from threedi_modelchecker.threedi_model.models import Control
from threedi_modelchecker.threedi_model.models import ControlGroup
from threedi_modelchecker.threedi_model.models import ControlMeasureGroup
from threedi_modelchecker.threedi_model.models import ControlMeasureMap
from threedi_modelchecker.threedi_model.models import ControlMemory
from threedi_modelchecker.threedi_model.models import ControlTable
from typing import Dict
from typing import List
from typing import Union


INFINITE_SIM_DURATION = int((timedelta(days=365) * 100).total_seconds())


def control_measure_map_to_measure_location(
    c_measure_map: ControlMeasureMap,
) -> MeasureLocation:
    # Connection nodes should be only option here.
    CONTENT_TYPE_MAP = {
        MeasureLocationContentTypes.v2_connection_nodes: "v2_connection_node"
    }

    return MeasureLocation(
        weight=str(round(float(c_measure_map.weight), 2)),
        content_type=CONTENT_TYPE_MAP[c_measure_map.object_type],
        content_pk=c_measure_map.object_id,
    )


def to_measure_specification(
    control: Union[ControlMemory, ControlTable],
    group: ControlGroup,
    locations: List[MeasureLocation],
) -> MeasureSpecification:
    VARIABLE_MAPPING = {
        MeasureVariables.waterlevel: "s1",
        MeasureVariables.volume: "vol1",
        MeasureVariables.discharge: "q",
        MeasureVariables.velocity: "u1",
    }

    # Use > as default for memory control
    operator = ">"
    if hasattr(control, "measure_operator"):
        operator = control.measure_operator.value

    return MeasureSpecification(
        name=group.name[:50] if group.name else "",
        variable=VARIABLE_MAPPING[control.measure_variable],
        locations=locations,
        operator=operator,
    )


TYPE_MAPPING = {ControlTableActionTypes.set_capacity: "set_pump_capacity"}
CAPACITY_FACTOR: float = 0.001


def parse_action_value(value) -> List[float]:
    # First clean the input:
    # - remove whitespace at the start & end (strip)
    # - replace 'internal' whitespace with ;
    # - replace commas with ;
    value = (";".join(value.strip().split())).replace(",", ";")
    # This could have yielded double semicolons (in case of <whitespace>;)
    while ";;" in value:
        value.replace(";;", ";")
    return [float(y) for y in value.split(";")]


def parse_action_table(table) -> List[List[float]]:
    """Parse an action table (for table controls)

    For table controls, the action table consists of multiple records, with
    2 or 3 values each, like:

    - "1.2;2.3#1.3;2.1#1.5;5.6"
    - "1.2;2.3;3.4#1.3;2.1;5.6#1.5;5.6;6.7"

    Sometimes, the values are separated by a space and not a semicolon (;). This
    function also accepts those inputs.
    """
    return [parse_action_value(x) for x in table.split("#")]


def to_table_control(
    control: Control,
    table_control: ControlTable,
    measure_specification: MeasureSpecification,
) -> TableStructureControl:

    action_type: str = TYPE_MAPPING.get(
        table_control.action_type, table_control.action_type.value
    )
    try:
        values = parse_action_table(table_control.action_table)
        if table_control.action_type is ControlTableActionTypes.set_capacity:
            values = [[x[0], x[1] * CAPACITY_FACTOR] for x in values]
    except (ValueError, TypeError):
        raise SchematisationError(
            f"Table control action_table incorrect format for v2_control_table.id = {table_control.id}"
        )

    try:
        control_start = int(control.start)
    except (ValueError, TypeError):
        control_start = 0

    try:
        control_end = int(control.end)
    except (ValueError, TypeError):
        control_end = INFINITE_SIM_DURATION

    return TableStructureControl(
        offset=control_start,
        duration=control_end - control_start,
        measure_specification=measure_specification,
        structure_id=table_control.target_id,
        structure_type=table_control.target_type.value,
        type=action_type,
        values=values,
    )


def to_memory_control(
    control: Control,
    memory_control: ControlMemory,
    measure_specification: MeasureSpecification,
) -> MemoryStructureControl:

    action_type: str = TYPE_MAPPING.get(
        memory_control.action_type, memory_control.action_type.value
    )

    try:
        value = parse_action_value(memory_control.action_value)
        if memory_control.action_type is ControlTableActionTypes.set_capacity:
            value = [value[0] * CAPACITY_FACTOR]
    except (ValueError, TypeError):
        raise SchematisationError(
            f"Memory control action_value incorrect format for v2_control_memory.id = {memory_control.id}"
        )

    try:
        control_start = int(control.start)
    except (ValueError, TypeError):
        control_start = 0

    try:
        control_end = int(control.end)
    except (ValueError, TypeError):
        control_end = INFINITE_SIM_DURATION

    return MemoryStructureControl(
        offset=control_start,
        duration=control_end - control_start,
        measure_specification=measure_specification,
        structure_id=memory_control.target_id,
        structure_type=memory_control.target_type.value,
        type=action_type,
        value=value,
        upper_threshold=memory_control.upper_threshold,
        lower_threshold=memory_control.lower_threshold,
        is_inverse=bool(memory_control.is_inverse),
        is_active=bool(memory_control.is_active),
    )


class StructureControlExtractor(object):
    def __init__(self, session: Session, control_group_id: int):
        self.session = session
        self._controls = None
        self._control_group_id = control_group_id

    def __initialize_controls(self):
        if self._controls is None:
            self._controls = {"table": [], "memory": []}
            table_lookup = dict(
                [
                    (x.id, x)
                    for x in Query(ControlTable).with_session(self.session).all()
                ]
            )
            memory_lookup = dict(
                [
                    (x.id, x)
                    for x in Query(ControlMemory).with_session(self.session).all()
                ]
            )
            maps_lookup = {}

            for map_item in Query([ControlMeasureMap]).with_session(self.session).all():
                if map_item.measure_group_id not in maps_lookup:
                    maps_lookup[map_item.measure_group_id] = []
                maps_lookup[map_item.measure_group_id].append(
                    control_measure_map_to_measure_location(map_item)
                )

            all_controls = (
                Query([Control, ControlGroup, ControlMeasureGroup])
                .join(ControlGroup, ControlMeasureGroup)
                .with_session(self.session)
                .filter(
                    Control.control_group_id == self._control_group_id,
                    ControlGroup.id == self._control_group_id,
                )
                .all()
            )

            for control, group, measuregroup in all_controls:
                control: Control
                maps: List[ControlMeasureGroup] = maps_lookup[measuregroup.id]

                api_control = None

                if control.control_type is ControlType.table:
                    table: ControlTable = table_lookup[control.control_id]
                    measure_spec = to_measure_specification(table, group, maps)
                    api_control = to_table_control(control, table, measure_spec)
                elif control.control_type is ControlType.memory:
                    memory: ControlMemory = memory_lookup[control.control_id]
                    measure_spec = to_measure_specification(memory, group, maps)
                    api_control = to_memory_control(control, memory, measure_spec)
                else:
                    raise SchematisationError(
                        f"Unknown control_type '{control.control_type.value}'"
                    )
                self._controls[control.control_type.value].append(api_control)

    def all_controls(self) -> StructureControls:
        self.__initialize_controls()
        return StructureControls(**self._controls)

    @property
    def memory_controls(self) -> List[MemoryStructureControl]:
        self.__initialize_controls()
        return self._controls["memory"]

    @property
    def table_controls(self) -> List[TableStructureControl]:
        self.__initialize_controls()
        return self._controls["table"]

    def as_list(self) -> List[Dict]:
        controls = []
        for control in self.all_controls():
            control_dict = control.to_dict()
            strip_dict_none_values(control_dict)
            controls.append(control_dict)
        return controls



expected_index_tables = [
    ("v2_2d_boundary_conditions", "the_geom"),
    ("v2_2d_lateral", "the_geom"),
    ("v2_calculation_point", "the_geom"),
    ("v2_channel", "the_geom"),
    ("v2_connected_pnt", "the_geom"),
    ("v2_connection_nodes", "the_geom"),
    ("v2_connection_nodes", "the_geom_linestring"),
    ("v2_cross_section_location", "the_geom"),
    ("v2_culvert", "the_geom"),
    ("v2_dem_average_area", "the_geom"),
    ("v2_floodfill", "the_geom"),
    ("v2_grid_refinement", "the_geom"),
    ("v2_grid_refinement_area", "the_geom"),
    ("v2_impervious_surface", "the_geom"),
    ("v2_initial_waterlevel", "the_geom"),
    ("v2_levee", "the_geom"),
    ("v2_obstacle", "the_geom"),
    ("v2_pumped_drainage_area", "the_geom"),
    ("v2_surface", "the_geom"),
    ("v2_windshielding", "the_geom"),
]



from threedi_modelchecker.checks.base import BaseCheck
from typing import NamedTuple


def print_errors(errors):
    """Simply prints all errors to stdout

    :param errors: iterator of BaseModelError
    """
    for error in errors:
        print(format_check_results(*error))


def export_to_file(errors, file):
    """Write errors to a new file, separated with newlines.

    File cannot be an already existing file.

    :param errors: iterator of BaseModelError
    :param file:
    :return: None
    :raise FileExistsError: if the file already exists
    """
    with open(file, "w") as f:
        for error in errors:
            f.write(format_check_results(*error) + "\n")


def format_check_results(check: BaseCheck, invalid_row: NamedTuple):
    OUTPUT_FORMAT = "{level}{error_code:04d} (id={row_id:d}) {description!s}"
    return OUTPUT_FORMAT.format(
        level=check.level.name[:1],
        error_code=check.error_code,
        row_id=invalid_row.id,
        description=check.description(),
    )



from ..threedi_model import constants
from ..threedi_model import models
from .base import BaseCheck
from .base import CheckLevel
from .geo_query import distance
from .geo_query import length
from .geo_query import transform
from dataclasses import dataclass
from geoalchemy2 import functions as geo_func
from sqlalchemy import func
from sqlalchemy import text
from sqlalchemy.orm import aliased
from sqlalchemy.orm import Query
from sqlalchemy.orm import Session
from typing import List
from typing import NamedTuple


class CrossSectionLocationCheck(BaseCheck):
    """Check if cross section locations are within {max_distance} of their channel."""

    def __init__(self, max_distance, *args, **kwargs):
        super().__init__(column=models.CrossSectionLocation.the_geom, *args, **kwargs)
        self.max_distance = max_distance

    def get_invalid(self, session):
        # get all channels with more than 1 cross section location
        return (
            self.to_check(session)
            .join(models.Channel)
            .filter(
                distance(models.CrossSectionLocation.the_geom, models.Channel.the_geom)
                > self.max_distance
            )
            .all()
        )

    def description(self):
        return (
            f"v2_cross_section_location.the_geom is invalid: the cross-section location "
            f"should be located on the channel geometry (tolerance = {self.max_distance} m)"
        )


class Use0DFlowCheck(BaseCheck):
    """Check that when use_0d_flow in global settings is configured to 1 or to
    2, there is at least one impervious surface or surfaces respectively.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(column=models.GlobalSetting.use_0d_inflow, *args, **kwargs)

    def to_check(self, session):
        """Return a Query object on which this check is applied"""
        return session.query(models.GlobalSetting).filter(
            models.GlobalSetting.use_0d_inflow != 0
        )

    def get_invalid(self, session):
        surface_count = session.query(func.count(models.Surface.id)).scalar()
        impervious_surface_count = session.query(
            func.count(models.ImperviousSurface.id)
        ).scalar()

        invalid_rows = []
        for row in self.to_check(session):
            if (
                row.use_0d_inflow == constants.InflowType.IMPERVIOUS_SURFACE
                and impervious_surface_count == 0
            ):
                invalid_rows.append(row)
            elif (
                row.use_0d_inflow == constants.InflowType.SURFACE and surface_count == 0
            ):
                invalid_rows.append(row)
            else:
                continue
        return invalid_rows

    def description(self):
        return (
            "When %s is used, there should exist at least one "
            "(impervious) surface." % self.column
        )


class ConnectionNodes(BaseCheck):
    """Check that all connection nodes are connected to at least one of the
    following objects:
    - Culvert
    - Channel
    - Pipe
    - Orifice
    - Pumpstation
    - Weir
    """

    def __init__(self, *args, **kwargs):
        super().__init__(column=models.ConnectionNode.id, *args, **kwargs)

    def get_invalid(self, session):
        raise NotImplementedError


class ConnectionNodesLength(BaseCheck):
    """Check that the distance between `start_node` and `end_node` is at least
    `min_distance`. The coords will be transformed into (the first entry) of
    GlobalSettings.epsg_code.
    """

    def __init__(self, start_node, end_node, min_distance: float, *args, **kwargs):
        """

        :param start_node: column name of the start node
        :param end_node: column name of the end node
        :param min_distance: minimum required distance between start and end node
        """
        super().__init__(*args, **kwargs)
        self.start_node = start_node
        self.end_node = end_node
        self.min_distance = min_distance

    def get_invalid(self, session):
        start_node = aliased(models.ConnectionNode)
        end_node = aliased(models.ConnectionNode)
        q = (
            Query(self.column.class_)
            .join(start_node, self.start_node)
            .join(end_node, self.end_node)
            .filter(
                distance(start_node.the_geom, end_node.the_geom) < self.min_distance
            )
        )
        return list(q.with_session(session).all())

    def description(self) -> str:
        return (
            f"The length of {self.table} is "
            f"very short (< {self.min_distance}). A length of at least 1.0 m is recommended."
        )


class ConnectionNodesDistance(BaseCheck):
    """Check that the distance between connection nodes is above a certain
    threshold
    """

    def __init__(self, minimum_distance: float, *args, **kwargs):
        """
        :param minimum_distance: threshold distance in degrees
        """
        super().__init__(
            column=models.ConnectionNode.id, level=CheckLevel.WARNING, *args, **kwargs
        )
        self.minimum_distance = minimum_distance

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        """
        The query makes use of the SpatialIndex so we won't have to calculate the
        distance between all connection nodes.
        """
        query = text(
            f"""SELECT *
               FROM v2_connection_nodes AS cn1, v2_connection_nodes AS cn2
               WHERE
                   distance(cn1.the_geom, cn2.the_geom, 1) < :min_distance
                   AND cn1.ROWID != cn2.ROWID
                   AND cn2.ROWID IN (
                     SELECT ROWID
                     FROM SpatialIndex
                     WHERE (
                       f_table_name = "v2_connection_nodes"
                       AND search_frame = Buffer(cn1.the_geom, {self.minimum_distance / 2})));
            """
        )
        results = (
            session.connection()
            .execute(query, min_distance=self.minimum_distance)
            .fetchall()
        )

        return results

    def description(self) -> str:
        return (
            f"The connection_node is within {self.minimum_distance} degrees of "
            f"another connection_node."
        )


class OpenChannelsWithNestedNewton(BaseCheck):
    """Checks whether the model has any closed cross-section in use when the
    NumericalSettings.use_of_nested_newton is turned off.

    See https://github.com/nens/threeditoolbox/issues/522
    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            column=models.CrossSectionDefinition.id,
            level=CheckLevel.WARNING,
            filters=Query(models.NumericalSettings)
            .filter(models.NumericalSettings.use_of_nested_newton == 0)
            .exists(),
            *args,
            **kwargs,
        )

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        definitions_in_use = self.to_check(session).filter(
            models.CrossSectionDefinition.id.in_(
                Query(models.CrossSectionLocation.definition_id).union_all(
                    Query(models.Pipe.cross_section_definition_id),
                    Query(models.Culvert.cross_section_definition_id),
                    Query(models.Weir.cross_section_definition_id),
                    Query(models.Orifice.cross_section_definition_id),
                )
            ),
        )

        # closed_rectangle, circle, and egg cross-section definitions are always closed:
        closed_definitions = definitions_in_use.filter(
            models.CrossSectionDefinition.shape.in_(
                [
                    constants.CrossSectionShape.CLOSED_RECTANGLE,
                    constants.CrossSectionShape.CIRCLE,
                    constants.CrossSectionShape.EGG,
                ]
            )
        )
        result = list(closed_definitions.with_session(session).all())

        # tabulated cross-section definitions are closed when the last element of 'width'
        # is zero
        tabulated_definitions = definitions_in_use.filter(
            models.CrossSectionDefinition.shape.in_(
                [
                    constants.CrossSectionShape.TABULATED_RECTANGLE,
                    constants.CrossSectionShape.TABULATED_TRAPEZIUM,
                ]
            )
        )
        for definition in tabulated_definitions.with_session(session).all():
            try:
                if float(definition.width.split(" ")[-1]) == 0.0:
                    # Closed channel
                    result.append(definition)
            except Exception:
                # Many things can go wrong, these are caught elsewhere
                pass
        return result

    def description(self) -> str:
        return (
            f"{self.column_name} has a closed cross section definition while "
            f"NumericalSettings.use_of_nested_newton is switched off. "
            f"This gives convergence issues. We recommend setting use_of_nested_newton = 1."
        )


class LinestringLocationCheck(BaseCheck):
    """Check that linestring geometry starts / ends are close to their connection nodes

    This allows for reversing the geometries. threedi-gridbuilder will reverse the geometries if
    that lowers the distance to the connection nodes.
    """

    def __init__(self, *args, **kwargs):
        self.max_distance = kwargs.pop("max_distance")
        super().__init__(*args, **kwargs)

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        start_node = aliased(models.ConnectionNode)
        end_node = aliased(models.ConnectionNode)

        tol = self.max_distance
        start_point = geo_func.ST_PointN(self.column, 1)
        end_point = geo_func.ST_PointN(self.column, geo_func.ST_NPoints(self.column))

        start_ok = distance(start_point, start_node.the_geom) <= tol
        end_ok = distance(end_point, end_node.the_geom) <= tol
        start_ok_if_reversed = distance(end_point, start_node.the_geom) <= tol
        end_ok_if_reversed = distance(start_point, end_node.the_geom) <= tol
        return (
            self.to_check(session)
            .join(start_node, start_node.id == self.table.c.connection_node_start_id)
            .join(end_node, end_node.id == self.table.c.connection_node_end_id)
            .filter(
                ~(start_ok & end_ok),
                ~(start_ok_if_reversed & end_ok_if_reversed),
            )
            .all()
        )

    def description(self) -> str:
        return f"{self.column_name} does not start or end at its connection node (tolerance = {self.max_distance} m)"


class BoundaryCondition1DObjectNumberCheck(BaseCheck):
    """Check that the number of connected objects to 1D boundary connections is 1."""

    def __init__(self, *args, **kwargs):
        super().__init__(
            column=models.BoundaryCondition1D.connection_node_id, *args, **kwargs
        )

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        invalid_ids = []
        for bc in self.to_check(session).all():
            total_objects = 0
            for table in [
                models.Channel,
                models.Pipe,
                models.Culvert,
                models.Orifice,
                models.Weir,
            ]:
                total_objects += (
                    session.query(table)
                    .filter(table.connection_node_start_id == bc.connection_node_id)
                    .count()
                )
                total_objects += (
                    session.query(table)
                    .filter(table.connection_node_end_id == bc.connection_node_id)
                    .count()
                )
            if total_objects != 1:
                invalid_ids.append(bc.id)

        return (
            self.to_check(session)
            .filter(models.BoundaryCondition1D.id.in_(invalid_ids))
            .all()
        )

    def description(self) -> str:
        return "1D boundary condition should be connected to exactly one object."


@dataclass
class IndexMissingRecord:
    id: int
    table_name: str
    column_name: str


class SpatialIndexCheck(BaseCheck):
    """Checks whether a spatial index is present and valid"""

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        result = session.execute(
            func.CheckSpatialIndex(self.column.table.name, self.column.name)
        ).scalar()
        if result == 1:
            return []
        else:
            return [
                IndexMissingRecord(
                    id=1,
                    table_name=self.column.table.name,
                    column_name=self.column.name,
                )
            ]

    def description(self) -> str:
        return f"{self.column_name} has no valid spatial index, which is required for some checks"


class PotentialBreachStartEndCheck(BaseCheck):
    """Check that a potential breach is exactly on or >=1 m from a linestring start/end."""

    def __init__(self, *args, **kwargs):
        self.min_distance = kwargs.pop("min_distance")

        super().__init__(*args, **kwargs)

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        linestring = models.Channel.the_geom
        tol = self.min_distance
        breach_point = func.Line_Locate_Point(
            transform(linestring), transform(geo_func.ST_PointN(self.column, 1))
        )
        dist_1 = breach_point * length(linestring)
        dist_2 = (1 - breach_point) * length(linestring)
        return (
            self.to_check(session)
            .join(models.Channel)
            .filter(((dist_1 > 0) & (dist_1 < tol)) | ((dist_2 > 0) & (dist_2 < tol)))
            .all()
        )

    def description(self) -> str:
        return f"{self.column_name} must be exactly on or >= {self.min_distance} m from a start or end channel vertex"


class PotentialBreachInterdistanceCheck(BaseCheck):
    """Check that a potential breaches are exactly on the same place or >=1 m apart."""

    def __init__(self, *args, **kwargs):
        self.min_distance = kwargs.pop("min_distance")
        assert "filters" not in kwargs

        super().__init__(*args, **kwargs)

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        # this query is hard to get performant; we do a hybrid sql / Python approach

        # First fetch the position of each potential breach per channel
        def get_position(point, linestring):
            breach_point = func.Line_Locate_Point(
                transform(linestring), transform(geo_func.ST_PointN(point, 1))
            )
            return (breach_point * length(linestring)).label("position")

        potential_breaches = sorted(
            session.query(
                self.table, get_position(self.column, models.Channel.the_geom)
            )
            .join(models.Channel)
            .all(),
            key=lambda x: (x.channel_id, x[-1]),
        )

        invalid = []
        prev_channel_id = -9999
        prev_position = -1.0
        for breach in potential_breaches:
            if breach.channel_id != prev_channel_id:
                prev_channel_id, prev_position = breach.channel_id, breach.position
                continue
            if breach.position == prev_position:
                continue
            if (breach.position - prev_position) <= self.min_distance:
                invalid.append(breach)
        return invalid

    def description(self) -> str:
        return f"{self.column_name} must be more than {self.min_distance} m apart (or exactly on the same position)"



VIEWS_TO_DELETE = ["v2_crosssection_view", "v2_pipe_map_view", "v2_imp_surface_view"]
ALL_VIEWS = {
    "v2_1d_lateral_view": {
        "definition": "SELECT a.rowid AS rowid, a.id AS id, a.connection_node_id AS connection_node_id, a.timeseries AS timeseries, b.the_geom FROM v2_1d_lateral a JOIN v2_connection_nodes b ON a.connection_node_id = b.id",
        "view_geometry": "the_geom",
        "view_rowid": "connection_node_id",
        "f_table_name": "v2_connection_nodes",
        "f_geometry_column": "the_geom",
    },
    "v2_1d_boundary_conditions_view": {
        "definition": "SELECT a.rowid AS rowid, a.id AS id, a.connection_node_id AS connection_node_id, a.boundary_type AS boundary_type, a.timeseries AS timeseries, b.the_geom FROM v2_1d_boundary_conditions a JOIN v2_connection_nodes b ON a.connection_node_id = b.id",
        "view_geometry": "the_geom",
        "view_rowid": "connection_node_id",
        "f_table_name": "v2_connection_nodes",
        "f_geometry_column": "the_geom",
    },
    "v2_cross_section_location_view": {
        "definition": "SELECT loc.rowid as rowid, loc.id as loc_id, loc.code as loc_code, loc.reference_level as loc_reference_level, loc.bank_level as loc_bank_level, loc.friction_type as loc_friction_type, loc.friction_value as loc_friction_value, loc.definition_id as loc_definition_id, loc.channel_id as loc_channel_id, loc.the_geom as the_geom, def.id as def_id, def.shape as def_shape, def.width as def_width, def.code as def_code, def.height as def_height FROM v2_cross_section_location loc, v2_cross_section_definition def WHERE loc.definition_id = def.id",
        "view_geometry": "the_geom",
        "view_rowid": "rowid",
        "f_table_name": "v2_cross_section_location",
        "f_geometry_column": "the_geom",
    },
    "v2_cross_section_view": {
        "definition": "SELECT def.rowid AS rowid, def.id AS def_id, def.shape AS def_shape, def.width AS def_width, def.height AS def_height, def.code AS def_code, l.id AS l_id, l.channel_id AS l_channel_id, l.definition_id AS l_definition_id, l.reference_level AS l_reference_level, l.friction_type AS l_friction_type, l.friction_value AS l_friction_value, l.bank_level AS l_bank_level, l.code AS l_code, l.the_geom AS the_geom, ch.id AS ch_id, ch.display_name AS ch_display_name, ch.code AS ch_code, ch.calculation_type AS ch_calculation_type, ch.dist_calc_points AS ch_dist_calc_points, ch.zoom_category AS ch_zoom_category, ch.connection_node_start_id AS ch_connection_node_start_id, ch.connection_node_end_id AS ch_connection_node_end_id FROM v2_cross_section_definition AS def , v2_cross_section_location AS l , v2_channel AS ch WHERE l.definition_id = def.id AND l.channel_id = ch.id",
        "view_geometry": "the_geom",
        "view_rowid": "rowid",
        "f_table_name": "v2_cross_section_location",
        "f_geometry_column": "the_geom",
    },
    "v2_culvert_view": {
        "definition": "SELECT cul.rowid AS rowid, cul.id AS cul_id, cul.display_name AS cul_display_name, cul.code AS cul_code, cul.calculation_type AS cul_calculation_type, cul.friction_value AS cul_friction_value, cul.friction_type AS cul_friction_type, cul.dist_calc_points AS cul_dist_calc_points, cul.zoom_category AS cul_zoom_category, cul.cross_section_definition_id AS cul_cross_section_definition_id, cul.discharge_coefficient_positive AS cul_discharge_coefficient_positive, cul.discharge_coefficient_negative AS cul_discharge_coefficient_negative, cul.invert_level_start_point AS cul_invert_level_start_point, cul.invert_level_end_point AS cul_invert_level_end_point, cul.the_geom AS the_geom, cul.connection_node_start_id AS cul_connection_node_start_id, cul.connection_node_end_id AS cul_connection_node_end_id, def.id AS def_id, def.shape AS def_shape, def.width AS def_width, def.height AS def_height, def.code AS def_code FROM v2_culvert AS cul , v2_cross_section_definition AS def WHERE cul.cross_section_definition_id = def.id",
        "view_geometry": "the_geom",
        "view_rowid": "rowid",
        "f_table_name": "v2_culvert",
        "f_geometry_column": "the_geom",
    },
    "v2_manhole_view": {
        "definition": "SELECT manh.rowid AS rowid, manh.id AS manh_id, manh.display_name AS manh_display_name, manh.code AS manh_code, manh.connection_node_id AS manh_connection_node_id, manh.shape AS manh_shape, manh.width AS manh_width, manh.length AS manh_length, manh.manhole_indicator AS manh_manhole_indicator, manh.calculation_type AS manh_calculation_type, manh.bottom_level AS manh_bottom_level, manh.surface_level AS manh_surface_level, manh.drain_level AS manh_drain_level, manh.sediment_level AS manh_sediment_level, manh.zoom_category AS manh_zoom_category, node.id AS node_id, node.storage_area AS node_storage_area, node.initial_waterlevel AS node_initial_waterlevel, node.code AS node_code, node.the_geom AS the_geom, node.the_geom_linestring AS node_the_geom_linestring FROM v2_manhole AS manh , v2_connection_nodes AS node WHERE manh.connection_node_id = node.id",
        "view_geometry": "the_geom",
        "view_rowid": "rowid",
        "f_table_name": "v2_connection_nodes",
        "f_geometry_column": "the_geom",
    },
    "v2_orifice_view": {
        "definition": "SELECT orf.rowid AS rowid, orf.id AS orf_id, orf.display_name AS orf_display_name, orf.code AS orf_code, orf.crest_level AS orf_crest_level, orf.sewerage AS orf_sewerage, orf.cross_section_definition_id AS orf_cross_section_definition_id, orf.friction_value AS orf_friction_value, orf.friction_type AS orf_friction_type, orf.discharge_coefficient_positive AS orf_discharge_coefficient_positive, orf.discharge_coefficient_negative AS orf_discharge_coefficient_negative, orf.zoom_category AS orf_zoom_category, orf.crest_type AS orf_crest_type, orf.connection_node_start_id AS orf_connection_node_start_id, orf.connection_node_end_id AS orf_connection_node_end_id, def.id AS def_id, def.shape AS def_shape, def.width AS def_width, def.height AS def_height, def.code AS def_code, MakeLine( start_node.the_geom, end_node.the_geom) AS the_geom FROM v2_orifice AS orf, v2_cross_section_definition AS def, v2_connection_nodes AS start_node, v2_connection_nodes AS end_node where orf.connection_node_start_id = start_node.id AND orf.connection_node_end_id = end_node.id AND orf.cross_section_definition_id = def.id",
        "view_geometry": "the_geom",
        "view_rowid": "rowid",
        "f_table_name": "v2_connection_nodes",
        "f_geometry_column": "the_geom_linestring",
    },
    "v2_pipe_view": {
        "definition": "SELECT pipe.rowid AS rowid, pipe.id AS pipe_id, pipe.display_name AS pipe_display_name, pipe.code AS pipe_code, pipe.profile_num AS pipe_profile_num, pipe.sewerage_type AS pipe_sewerage_type, pipe.calculation_type AS pipe_calculation_type, pipe.invert_level_start_point AS pipe_invert_level_start_point, pipe.invert_level_end_point AS pipe_invert_level_end_point, pipe.cross_section_definition_id AS pipe_cross_section_definition_id, pipe.friction_value AS pipe_friction_value, pipe.friction_type AS pipe_friction_type, pipe.dist_calc_points AS pipe_dist_calc_points, pipe.material AS pipe_material, pipe.original_length AS pipe_original_length, pipe.zoom_category AS pipe_zoom_category, pipe.connection_node_start_id AS pipe_connection_node_start_id, pipe.connection_node_end_id AS pipe_connection_node_end_id, def.id AS def_id, def.shape AS def_shape, def.width AS def_width, def.height AS def_height, def.code AS def_code, MakeLine( start_node.the_geom, end_node.the_geom) AS the_geom FROM v2_pipe AS pipe , v2_cross_section_definition AS def , v2_connection_nodes AS start_node , v2_connection_nodes AS end_node WHERE pipe.connection_node_start_id = start_node.id AND pipe.connection_node_end_id = end_node.id AND pipe.cross_section_definition_id = def.id",
        "view_geometry": "the_geom",
        "view_rowid": "rowid",
        "f_table_name": "v2_connection_nodes",
        "f_geometry_column": "the_geom_linestring",
    },
    "v2_pumpstation_point_view": {
        "definition": "SELECT a.rowid AS rowid, a.id AS pump_id, a.display_name, a.code, a.classification, a.sewerage, a.start_level, a.lower_stop_level, a.upper_stop_level, a.capacity, a.zoom_category, a.connection_node_start_id, a.connection_node_end_id, a.type, b.id AS connection_node_id, b.storage_area, b.the_geom FROM v2_pumpstation a JOIN v2_connection_nodes b ON a.connection_node_start_id = b.id",
        "view_geometry": "the_geom",
        "view_rowid": "connection_node_start_id",
        "f_table_name": "v2_connection_nodes",
        "f_geometry_column": "the_geom",
    },
    "v2_pumpstation_view": {
        "definition": "SELECT pump.rowid AS rowid, pump.id AS pump_id, pump.display_name AS pump_display_name, pump.code AS pump_code, pump.classification AS pump_classification, pump.type AS pump_type, pump.sewerage AS pump_sewerage, pump.start_level AS pump_start_level, pump.lower_stop_level AS pump_lower_stop_level, pump.upper_stop_level AS pump_upper_stop_level, pump.capacity AS pump_capacity, pump.zoom_category AS pump_zoom_category, pump.connection_node_start_id AS pump_connection_node_start_id, pump.connection_node_end_id AS pump_connection_node_end_id, MakeLine( start_node.the_geom, end_node.the_geom ) AS the_geom FROM v2_pumpstation AS pump , v2_connection_nodes AS start_node , v2_connection_nodes AS end_node WHERE pump.connection_node_start_id = start_node.id AND pump.connection_node_end_id = end_node.id",
        "view_geometry": "the_geom",
        "view_rowid": "rowid",
        "f_table_name": "v2_connection_nodes",
        "f_geometry_column": "the_geom_linestring",
    },
    "v2_weir_view": {
        "definition": "SELECT weir.rowid AS rowid, weir.id AS weir_id, weir.display_name AS weir_display_name, weir.code AS weir_code, weir.crest_level AS weir_crest_level, weir.crest_type AS weir_crest_type, weir.cross_section_definition_id AS weir_cross_section_definition_id, weir.sewerage AS weir_sewerage, weir.discharge_coefficient_positive AS weir_discharge_coefficient_positive, weir.discharge_coefficient_negative AS weir_discharge_coefficient_negative, weir.external AS weir_external, weir.zoom_category AS weir_zoom_category, weir.friction_value AS weir_friction_value, weir.friction_type AS weir_friction_type, weir.connection_node_start_id AS weir_connection_node_start_id, weir.connection_node_end_id AS weir_connection_node_end_id, def.id AS def_id, def.shape AS def_shape, def.width AS def_width, def.height AS def_height, def.code AS def_code, MakeLine( start_node.the_geom, end_node.the_geom) AS the_geom FROM v2_weir AS weir , v2_cross_section_definition AS def , v2_connection_nodes AS start_node , v2_connection_nodes AS end_node WHERE weir.connection_node_start_id = start_node.id AND weir.connection_node_end_id = end_node.id AND weir.cross_section_definition_id = def.id",
        "view_geometry": "the_geom",
        "view_rowid": "rowid",
        "f_table_name": "v2_connection_nodes",
        "f_geometry_column": "the_geom_linestring",
    },
}



"""fix schema constraints

Revision ID: 0207
Revises: 0206
Create Date: 2022-05-18 10:15:20.851968

"""
from alembic import op

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "0207"
down_revision = "0206"
branch_labels = None
depends_on = None


MIGRATION_QUERIES = """
DELETE FROM v2_grid_refinement WHERE (refinement_level IS NULL) OR (the_geom IS NULL);

DELETE FROM v2_grid_refinement_area WHERE (refinement_level IS NULL) OR (the_geom IS NULL);

UPDATE v2_orifice SET sewerage = NULL WHERE sewerage NOT IN (0, 1);
UPDATE v2_pumpstation SET sewerage = NULL WHERE sewerage NOT IN (0, 1);
UPDATE v2_weir SET sewerage = NULL WHERE sewerage NOT IN (0, 1);

UPDATE v2_weir SET external = NULL WHERE external NOT IN (0, 1);

DROP TRIGGER IF EXISTS ggi_v2_connection_nodes_the_geom;
DROP TRIGGER IF EXISTS ggu_v2_connection_nodes_the_geom;
DROP TRIGGER IF EXISTS ggi_v2_connection_nodes_the_geom_linestring;
DROP TRIGGER IF EXISTS ggu_v2_connection_nodes_the_geom_linestring;
UPDATE v2_connection_nodes SET the_geom_linestring = NULL;
"""


def upgrade():
    # First alter the tables to nullable TEXT, because we need to be able to
    # accept any value
    with op.batch_alter_table("v2_orifice") as batch_op:
        batch_op.alter_column("sewerage", nullable=True, type_=sa.TEXT)

    with op.batch_alter_table("v2_pumpstation") as batch_op:
        batch_op.alter_column("sewerage", nullable=True, type_=sa.TEXT)

    with op.batch_alter_table("v2_weir") as batch_op:
        batch_op.alter_column("sewerage", nullable=True, type_=sa.TEXT)
        batch_op.alter_column("external", nullable=True, type_=sa.TEXT)

    for q in MIGRATION_QUERIES.split(";"):
        op.execute(q)

    # After the data migration; alter the tables to nullable BOOLEAN
    with op.batch_alter_table("v2_orifice") as batch_op:
        batch_op.alter_column("sewerage", nullable=True, type_=sa.BOOLEAN)

    with op.batch_alter_table("v2_pumpstation") as batch_op:
        batch_op.alter_column("sewerage", nullable=True, type_=sa.BOOLEAN)

    with op.batch_alter_table("v2_weir") as batch_op:
        batch_op.alter_column("sewerage", nullable=True, type_=sa.BOOLEAN)
        batch_op.alter_column("external", nullable=True, type_=sa.BOOLEAN)


def downgrade():
    pass



from .exceptions import TemplateValidationError
from .exceptions import TemplateValidationTimeoutError
from dataclasses import dataclass
from dataclasses import fields
from dataclasses import InitVar
from enum import Enum
from io import BytesIO
from threedi_api_client.aio.files import upload_fileobj
from threedi_api_client.openapi.models import AggregationSettings
from threedi_api_client.openapi.models import FileLateral
from threedi_api_client.openapi.models import InitialWaterlevel
from threedi_api_client.openapi.models import Lateral
from threedi_api_client.openapi.models import MeasureLocation
from threedi_api_client.openapi.models import MeasureSpecification
from threedi_api_client.openapi.models import MemoryStructureControl
from threedi_api_client.openapi.models import NumericalSettings
from threedi_api_client.openapi.models import PhysicalSettings
from threedi_api_client.openapi.models import Simulation
from threedi_api_client.openapi.models import TableStructureControl
from threedi_api_client.openapi.models import Template
from threedi_api_client.openapi.models import TimeStepSettings
from threedi_api_client.openapi.models import UploadEventFile
from threedi_api_client.openapi.models.file_boundary_condition import (
    FileBoundaryCondition,
)
from threedi_api_client.openapi.models.file_structure_control import (
    FileStructureControl,
)
from threedi_api_client.openapi.models.ground_water_level import (
    GroundWaterLevel,
)
from threedi_api_client.openapi.models.ground_water_raster import (
    GroundWaterRaster,
)
from threedi_api_client.openapi.models.one_d_water_level import OneDWaterLevel
from threedi_api_client.openapi.models.one_d_water_level_file import (
    OneDWaterLevelFile,
)
from threedi_api_client.openapi.models.two_d_water_level import TwoDWaterLevel
from threedi_api_client.openapi.models.two_d_water_raster import (
    TwoDWaterRaster,
)
from threedi_api_client.versions import V3BetaApi
from threedi_modelchecker.simulation_templates.utils import (
    strip_dict_none_values,
)
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from uuid import uuid4

import asyncio
import json


__all__ = [
    "ValidationStatus",
    "GlobalSettingOption",
    "InitialWaterlevels",
    "StructureControls",
    "Settings",
    "Events",
    "SimulationTemplate",
]


class ValidationStatus(Enum):
    processing: str = "processing"
    valid: str = "valid"
    invalid: str = "invalid"


@dataclass
class GlobalSettingOption:
    """
    Derived global setting option from v2_global_settings table
    """

    id: int  # v2_global_settings.id
    name: str


class AsyncBytesIO:
    """
    Simple wrapper class to make BytesIO async
    """

    def __init__(self, bytes_io: BytesIO):
        self._bytes_io = bytes_io

    async def seek(self, *args, **kwargs):
        return self._bytes_io.seek(*args, **kwargs)

    async def read(self, *args, **kwargs):
        return self._bytes_io.read(*args, **kwargs)


async def get_upload_instance(
    api_func, simulation_pk: int, filename: str
) -> Optional[Any]:
    """
    Get the uploaded file instance based on filename

    api_func should be an async API function to list a
    file upload instance
    """
    offset: int = 0
    limit: int = 10

    found = None
    next: bool = True

    while not found and next:
        res = await api_func(simulation_pk=simulation_pk, offset=offset, limit=limit)
        results = [x for x in res.results if x.file.filename == filename]

        if results:
            found = results[0]
            break

        # Make sure AsyncMock used in unit-tests are not evaluated as true
        next = res.next is not None and isinstance(res.next, str) and len(res.next) != 0
        offset += 10
        limit += 10

    return found


def openapi_to_dict(value: Any):
    """
    Convert openapi object to Dict
    """
    if hasattr(value, "openapi_types") and hasattr(value, "to_dict"):
        value = value.to_dict()
        strip_dict_none_values(value)
    return value


class AsDictMixin:
    def as_dict(self) -> Dict:
        """
        Convert fields to dictionary
        """
        rt = {}
        for field_name in [x.name for x in fields(self)]:
            value = getattr(self, field_name)
            if isinstance(value, AsDictMixin):
                value = value.as_dict()
            elif isinstance(value, list):
                value = [openapi_to_dict(x) for x in value]
            else:
                value = openapi_to_dict(value)
            rt[field_name] = value
        return rt


@dataclass
class InitialWaterlevels(AsDictMixin):
    constant_2d: Optional[TwoDWaterLevel] = None
    constant_1d: Optional[OneDWaterLevel] = None
    constant_gw: Optional[GroundWaterLevel] = None
    waterlevel_1d_file: Optional[OneDWaterLevelFile] = None
    raster_2d: Optional[TwoDWaterRaster] = None
    raster_gw: Optional[GroundWaterRaster] = None

    _validation_status: InitVar[ValidationStatus] = ValidationStatus.processing

    async def is_valid_in_api(self, client: V3BetaApi) -> ValidationStatus:
        return self._validation_status

    async def save_to_api(self, client: V3BetaApi, simulation: Simulation):
        """
        :param: client = ThreediApi(async=True)

        Save initial waterlevels to the API on the given simulation
        """
        if self._validation_status == ValidationStatus.valid:
            # Already saved
            return

        self._validation_status == ValidationStatus.processing

        tasks = []
        if self.constant_1d is not None:
            tasks.append(
                client.simulations_initial1d_water_level_constant_create(
                    simulation_pk=simulation.id, data=self.constant_1d
                )
            )
        if self.constant_2d is not None:
            tasks.append(
                client.simulations_initial2d_water_level_constant_create(
                    simulation_pk=simulation.id, data=self.constant_2d
                )
            )
        if self.constant_gw is not None:
            tasks.append(
                client.simulations_initial_groundwater_level_constant_create(
                    simulation_pk=simulation.id, data=self.constant_gw
                )
            )
        initial_waterlevels: List[InitialWaterlevel] = []
        rasters_lookup: Dict = {}

        # TODO: support user uploaded 1d/2d/gw initial waterlevel file/rasters to threedimodels.

        if (
            self.raster_2d is not None
            or self.raster_gw is not None
            or self.waterlevel_1d_file is not None
        ):
            # Fetch initial waterlevels
            initial_waterlevels = (
                await client.threedimodels_initial_waterlevels_list(
                    simulation.threedimodel_id
                )
            ).results
            rasters_lookup = dict(
                [
                    (x.id, x)
                    for x in (
                        await client.threedimodels_rasters_list(
                            simulation.threedimodel_id
                        )
                    ).results
                ]
            )

            for initial_waterlevel in initial_waterlevels:
                initial_waterlevel._raster = rasters_lookup.get(
                    initial_waterlevel.source_raster_id, None
                )

        if self.raster_2d is not None:
            # use first found initial waterlevel resource with:
            #   - dimension = "two_d"
            #   - source_raster.type = "initial_waterlevel_file"
            found = [
                x
                for x in initial_waterlevels
                if x.dimension == "two_d"
                and x._raster
                and x._raster.type == "initial_waterlevel_file"
            ]
            if not found:
                raise TemplateValidationError(
                    "Could not find aggregation file for 2D initial waterlevel raster"
                )
            self.raster_2d.initial_waterlevel = found[0].id
            tasks.append(
                client.simulations_initial2d_water_level_raster_create(
                    simulation_pk=simulation.id, data=self.raster_2d
                )
            )

        if self.raster_gw is not None:
            # use first found initial waterlevel resource with:
            #   - dimension = "two_d"
            #   - source_raster.type = "initial_groundwater_level_file"
            found = [
                x
                for x in initial_waterlevels
                if x.dimension == "two_d"
                and x._raster
                and x._raster.type == "initial_groundwater_level_file"
            ]
            if not found:
                raise TemplateValidationError(
                    "Could not find aggregation file for initial groundwaterlevel raster"
                )
            self.raster_gw.initial_waterlevel = found[0].id
            tasks.append(
                client.simulations_initial_groundwater_level_raster_create(
                    simulation_pk=simulation.id, data=self.raster_gw
                )
            )

        if self.waterlevel_1d_file is not None:
            # use first found initial waterlevel resource with:
            #   - dimension = "one_d"
            found = [x for x in initial_waterlevels if x.dimension == "one_d"]
            if not found:
                raise TemplateValidationError(
                    "Could not find aggregation file for 1D initial waterlevel from file"
                )

            self.waterlevel_1d_file.initial_waterlevel = found[0].id
            tasks.append(
                client.simulations_initial1d_water_level_file_create(
                    simulation_pk=simulation.id, data=self.waterlevel_1d_file
                )
            )

        if tasks:
            await asyncio.gather(*tasks)

        self._validation_status = ValidationStatus.valid

    @classmethod
    def from_dict(cls, dict: Dict) -> "InitialWaterlevels":
        map = {
            "constant_2d": TwoDWaterLevel,
            "constant_1d": OneDWaterLevel,
            "constant_gw": GroundWaterLevel,
            "waterlevel_1d_file": OneDWaterLevelFile,
            "raster_2d": TwoDWaterRaster,
            "raster_gw": GroundWaterRaster,
        }

        data = {}
        for key, klass in map.items():
            data[key] = None if dict[key] is None else klass(**dict[key])

        return InitialWaterlevels(**data)


@dataclass
class StructureControls(AsDictMixin):
    memory: List[MemoryStructureControl]
    table: List[TableStructureControl]

    _controls_upload: InitVar[Optional[FileStructureControl]] = None
    _validation_status: InitVar[Optional[ValidationStatus]] = None
    _simulation: InitVar[Optional[Simulation]] = None

    async def is_valid_in_api(self, client: V3BetaApi) -> ValidationStatus:
        """
        Return ValidationStatus of uploaded controls file
        """
        if self._controls_upload is None:
            return ValidationStatus.valid

        if ValidationStatus[self._controls_upload.state] == ValidationStatus.processing:
            # Refresh from API
            self._controls_upload: FileStructureControl = (
                await client.simulations_events_structure_control_file_read(
                    id=self._controls_upload.id, simulation_pk=self._simulation.id
                )
            )

        validation_status: ValidationStatus = ValidationStatus[
            self._controls_upload.state
        ]

        if validation_status == ValidationStatus.invalid:
            raise TemplateValidationError(
                f"Provided structure controls could not be validated succesfully, {self._controls_upload.state_detail}"
            )

        return validation_status

    @classmethod
    def from_dict(cls, dict: Dict) -> "StructureControls":
        def convert_measure_specs(data: Dict):
            """
            Convert measure specification on tables/memory from dict to OpenAPI models
            """
            rt = {}
            for k, v in data.items():
                if k == "measure_specification":
                    v = MeasureSpecification(**data[k])
                    v.locations = [MeasureLocation(**x) for x in data[k]["locations"]]

                rt[k] = v
            return rt

        return StructureControls(
            memory=[
                MemoryStructureControl(**convert_measure_specs(x))
                for x in dict["memory"]
            ],
            table=[
                TableStructureControl(**convert_measure_specs(x)) for x in dict["table"]
            ],
        )

    @property
    def has_controls(self) -> bool:
        return len(self.memory) + len(self.table) > 0

    async def save_to_api(self, client: V3BetaApi, simulation: Simulation):
        """
        :param: client = ThreediApi(async=True)

        Save structure controls to API on the given simulation

        Saves them individual if the total count <= 30, else saves
        them using file upload.
        """
        self._simulation = simulation

        if not self.has_controls:
            return

        data: Dict = {"memory": [], "table": []}
        for memory_control in self.memory:
            data["memory"].append(openapi_to_dict(memory_control))
        for table_control in self.table:
            data["table"].append(openapi_to_dict(table_control))
        data["timed"] = []  # not supported via sqlite

        filename: str = f"controls{uuid4().hex[0:8]}.json"

        upload: UploadEventFile = (
            await client.simulations_events_structure_control_file_create(
                simulation_pk=simulation.id,
                data=UploadEventFile(filename=filename, offset=0),
            )
        )

        await upload_fileobj(
            upload.put_url, AsyncBytesIO(BytesIO(json.dumps(data).encode()))
        )

        # Try to find structure control resource
        controls_upload: Optional[FileStructureControl] = await get_upload_instance(
            client.simulations_events_structure_control_file_list,
            simulation.id,
            filename,
        )

        if controls_upload is None:
            raise TemplateValidationError(
                "Could not find uploaded structure controls file resource"
            )
        self._controls_upload = controls_upload


@dataclass
class Settings(AsDictMixin):
    numerical: NumericalSettings
    physical: PhysicalSettings
    timestep: TimeStepSettings
    aggregations: List[AggregationSettings]
    _validation_status: InitVar[ValidationStatus] = ValidationStatus.processing

    async def is_valid_in_api(self, client: V3BetaApi) -> ValidationStatus:
        return self._validation_status

    @classmethod
    def from_dict(cls, dict: Dict) -> "Settings":
        return Settings(
            numerical=NumericalSettings(**dict["numerical"]),
            physical=PhysicalSettings(**dict["physical"]),
            timestep=TimeStepSettings(**dict["timestep"]),
            aggregations=[AggregationSettings(**x) for x in dict["aggregations"]],
        )

    async def save_to_api(self, client: V3BetaApi, simulation: Simulation):
        """
        :param: client = ThreediApi(async=True)

        Save settings to API on given simulation
        """
        if self._validation_status == ValidationStatus.valid:
            # Already saved
            return

        self._validation_status = ValidationStatus.processing

        tasks = [
            client.simulations_settings_numerical_create(
                simulation_pk=simulation.id, data=self.numerical
            ),
            client.simulations_settings_physical_create(
                simulation_pk=simulation.id, data=self.physical
            ),
            client.simulations_settings_time_step_create(
                simulation_pk=simulation.id, data=self.timestep
            ),
        ]
        for aggregation in self.aggregations:
            tasks.append(
                client.simulations_settings_aggregation_create(
                    simulation_pk=simulation.id, data=aggregation
                )
            )

        # Add aggregations
        await asyncio.gather(*tasks)

        self._validation_status = ValidationStatus.valid


@dataclass
class Events(AsDictMixin):
    laterals: List[Lateral]
    dwf_laterals: List[Dict]
    boundaries: List[Dict]
    structure_controls: StructureControls

    _validation_status: InitVar[Optional[ValidationStatus]] = None
    _lateral_upload: InitVar[Optional[FileLateral]] = None
    _dwf_lateral_upload: InitVar[Optional[FileLateral]] = None
    _boundary_upload: InitVar[Optional[FileBoundaryCondition]] = None

    # Simulation used to store events
    _simulation: InitVar[Optional[Simulation]] = None

    async def _is_laterals_valid_in_api(self, client: V3BetaApi) -> ValidationStatus:
        """
        Return ValidationStatus of uploaded lateral file
        """
        if self._lateral_upload is None:
            return ValidationStatus.valid

        if ValidationStatus[self._lateral_upload.state] == ValidationStatus.processing:
            # Refresh from API
            self._lateral_upload: FileLateral = (
                await client.simulations_events_lateral_file_read(
                    id=self._lateral_upload.id, simulation_pk=self._simulation.id
                )
            )

        validation_status: ValidationStatus = ValidationStatus[
            self._lateral_upload.state
        ]

        if validation_status == ValidationStatus.invalid:
            raise TemplateValidationError(
                f"Provided laterals could not be validated succesfully, {self._lateral_upload.state_detail}"
            )

        return validation_status

    async def _is_dwf_laterals_valid_in_api(
        self, client: V3BetaApi
    ) -> ValidationStatus:
        """
        Return ValidationStatus of uploaded DWF lateral file
        """
        if self._dwf_lateral_upload is None:
            return ValidationStatus.valid

        if (
            ValidationStatus[self._dwf_lateral_upload.state]
            == ValidationStatus.processing
        ):
            # Refresh from API
            self._dwf_lateral_upload: FileLateral = (
                await client.simulations_events_lateral_file_read(
                    id=self._dwf_lateral_upload.id, simulation_pk=self._simulation.id
                )
            )

        validation_status = ValidationStatus[self._dwf_lateral_upload.state]

        if validation_status == ValidationStatus.invalid:
            raise TemplateValidationError(
                f"Provided laterals could not be validated succesfully, {self._dwf_lateral_upload.state_detail}"
            )

        return validation_status

    async def _is_boundaries_valid_in_api(self, client: V3BetaApi) -> ValidationStatus:
        """
        Return ValidationStatus of uploaded boundary file
        """
        if self._boundary_upload is None:
            return ValidationStatus.valid

        if ValidationStatus[self._boundary_upload.state] == ValidationStatus.processing:
            # Refresh from API
            self._boundary_upload: FileBoundaryCondition = (
                await client.simulations_events_boundaryconditions_file_read(
                    id=self._boundary_upload.id, simulation_pk=self._simulation.id
                )
            )

        validation_status: ValidationStatus = ValidationStatus[
            self._boundary_upload.state
        ]

        if validation_status == ValidationStatus.invalid:
            raise TemplateValidationError(
                f"Provided boundary conditions could not be validated succesfully, {self._boundary_upload.state_detail}"
            )

        return validation_status

    async def is_valid_in_api(self, client: V3BetaApi) -> ValidationStatus:
        """
        Returns ValidationStatus (processing or valid) or raises TemplateValidationError in case
        something is invalid
        """
        if await self._is_laterals_valid_in_api(client) == ValidationStatus.processing:
            return ValidationStatus.processing

        if (
            await self._is_dwf_laterals_valid_in_api(client)
            == ValidationStatus.processing
        ):
            return ValidationStatus.processing

        if (
            await self._is_boundaries_valid_in_api(client)
            == ValidationStatus.processing
        ):
            return ValidationStatus.processing

        if (
            await self.structure_controls.is_valid_in_api(client)
            == ValidationStatus.processing
        ):
            return ValidationStatus.processing

        # If no errors have been raised, the validation is succesfully
        return ValidationStatus.valid

    async def save_to_api(self, client: V3BetaApi, simulation: Simulation):
        """
        :param: client = ThreediApi(async=True)

        Save events to API on the given simulation
        """
        self._simulation = simulation

        tasks = [
            # Laterals
            self.save_laterals_to_api(client, simulation),
            self.save_dwf_laterals_to_api(client, simulation),
            # Boundaries
            self.save_boundaries_to_api(client, simulation),
            # Structure controls
            self.structure_controls.save_to_api(client, simulation),
        ]
        await asyncio.gather(*tasks)

    async def save_laterals_to_api(self, client: V3BetaApi, simulation: Simulation):
        """
        Save laterals to API on the given simulation as file upload.
        """
        if self._simulation is None:
            self._simulation = simulation

        if len(self.laterals) == 0:
            return

        data = [openapi_to_dict(x) for x in self.laterals]
        data = AsyncBytesIO(BytesIO(json.dumps(data).encode()))

        filename: str = f"laterals_{uuid4().hex[:8]}.json"

        upload: UploadEventFile = await client.simulations_events_lateral_file_create(
            simulation_pk=simulation.id,
            data=UploadEventFile(filename=filename, offset=0),
        )
        await upload_fileobj(upload.put_url, data)

        # Try to find lateral uploaded resource
        lateral_upload: Optional[FileLateral] = await get_upload_instance(
            client.simulations_events_lateral_file_list, simulation.id, filename
        )

        if lateral_upload is None:
            raise TemplateValidationError(
                "Could not find uploaded lateral file resource"
            )
        self._lateral_upload = lateral_upload

    async def save_dwf_laterals_to_api(self, client: V3BetaApi, simulation: Simulation):
        """
        Save Dry Weather Flow (DWF) laterals to API on the given simulation as file upload.
        """
        if self._simulation is None:
            self._simulation = simulation

        if len(self.dwf_laterals) == 0:
            return

        data = AsyncBytesIO(BytesIO(json.dumps(self.dwf_laterals).encode()))
        filename: str = f"dwf_laterals_{uuid4().hex[:8]}.json"

        upload: UploadEventFile = await client.simulations_events_lateral_file_create(
            simulation_pk=simulation.id,
            data=UploadEventFile(filename=filename, offset=0, periodic="daily"),
        )
        await upload_fileobj(upload.put_url, data)

        # Try to find lateral uploaded resource
        dwf_lateral_upload: Optional[FileLateral] = await get_upload_instance(
            client.simulations_events_lateral_file_list, simulation.id, filename
        )

        if dwf_lateral_upload is None:
            raise TemplateValidationError(
                "Could not find uploaded DWF lateral file resource"
            )
        self._dwf_lateral_upload = dwf_lateral_upload

    async def save_boundaries_to_api(self, client: V3BetaApi, simulation: Simulation):
        """
        Save boundary to API on the given simulation
        """
        if self._simulation is None:
            self._simulation = simulation

        if len(self.boundaries) == 0:
            return

        data = [openapi_to_dict(x) for x in self.boundaries]
        data = AsyncBytesIO(BytesIO(json.dumps(data).encode()))

        filename: str = f"boundaries_{uuid4().hex[:8]}.json"

        upload: UploadEventFile = (
            await client.simulations_events_boundaryconditions_file_create(
                simulation_pk=simulation.id,
                data=UploadEventFile(filename=filename, offset=0),
            )
        )
        await upload_fileobj(upload.put_url, data)

        # Try to find boundary uploaded resource
        boundary_upload: Optional[FileBoundaryCondition] = await get_upload_instance(
            client.simulations_events_boundaryconditions_file_list,
            simulation.id,
            filename,
        )

        if boundary_upload is None:
            raise TemplateValidationError(
                "Could not find uploaded boundary file resource"
            )

        self._boundary_upload = boundary_upload

    @classmethod
    def from_dict(cls, dict: Dict) -> "Events":
        return Events(
            laterals=[Lateral(**x) for x in dict["laterals"]],
            boundaries=dict["boundaries"],
            structure_controls=StructureControls.from_dict(dict["structure_controls"]),
        )


@dataclass
class SimulationTemplate(AsDictMixin):
    settings: Settings
    events: Events
    initial_waterlevels: InitialWaterlevels

    # if saved to api, the Simulation the data is saved to
    _simulation: InitVar[Optional[Simulation]] = None

    async def is_valid_in_api(self, client: V3BetaApi) -> ValidationStatus:
        """
        Check if all data for this SimulationTemplate has been saved correctly in the API

        returns:
            ValidationStatus.valid if everything is ok
            ValidationStatus.processing if still validating (or nothing saved)

        raises: TemplateValidationError in case of invalid data.
        """
        if await self.settings.is_valid_in_api(client) == ValidationStatus.processing:
            return ValidationStatus.processing

        if await self.events.is_valid_in_api(client) == ValidationStatus.processing:
            return ValidationStatus.processing

        if (
            await self.initial_waterlevels.is_valid_in_api(client)
            == ValidationStatus.processing
        ):
            return ValidationStatus.processing

        # All are valid (none is processing and no TemplateValidationError)
        return ValidationStatus.valid

    async def save_to_api(
        self,
        client: V3BetaApi,
        template_name: str,
        simulation: Simulation,
        timeout: int = 300,
    ) -> Template:

        if simulation.id is None:
            # Save Simulation
            simulation: Simulation = await client.simulations_create(data=simulation)

        self._simulation = simulation

        # Conditonally save things to simulation in the API,
        # if not done already
        await self.settings.save_to_api(client, simulation)
        await self.events.save_to_api(client, simulation)
        await self.initial_waterlevels.save_to_api(client, simulation)

        # Check validity of everything
        async def wait_for_validation(client: V3BetaApi) -> None:
            # increase sleep time, in steps to 8 sec
            SLEEPTIME: int = [2, 3, 5, 8]
            index: int = 0

            while await self.is_valid_in_api(client) == ValidationStatus.processing:
                if index < len(SLEEPTIME) - 2:
                    index += 1
                await asyncio.sleep(SLEEPTIME[index])

        try:
            await asyncio.wait_for(wait_for_validation(client), timeout=timeout)
        except TimeoutError:
            raise TemplateValidationTimeoutError(
                "Template validating timed out, please try again later."
            )

        # All things should be valid now, save template
        template: Template = await client.simulation_templates_create(
            data=Template(simulation=simulation.id, name=template_name)
        )
        return template

    @classmethod
    def from_dict(cls, dict: Dict) -> "SimulationTemplate":
        return SimulationTemplate(
            settings=Settings.from_dict(dict["settings"]),
            events=Events.from_dict(dict["events"]),
            initial_waterlevels=InitialWaterlevels.from_dict(
                dict["initial_waterlevels"]
            ),
        )



from ..threedi_model import models
from .base import BaseCheck
from sqlalchemy.orm import Query


class CrossSectionBaseCheck(BaseCheck):
    """Base class for all cross section definition checks."""

    def __init__(self, column, *args, **kwargs):
        self.shapes = kwargs.pop("shapes", None)
        super().__init__(column, *args, **kwargs)

    @property
    def shape_msg(self):
        return {x.value for x in self.shapes}

    def to_check(self, session):
        qs = super().to_check(session)
        if self.shapes is not None:
            qs = qs.filter(models.CrossSectionDefinition.shape.in_(self.shapes))
        return qs.filter(
            models.CrossSectionDefinition.id.in_(
                Query(models.CrossSectionLocation.definition_id).union_all(
                    Query(models.Pipe.cross_section_definition_id),
                    Query(models.Culvert.cross_section_definition_id),
                    Query(models.Weir.cross_section_definition_id),
                    Query(models.Orifice.cross_section_definition_id),
                )
            )
        )


class CrossSectionNullCheck(CrossSectionBaseCheck):
    """Check if width / height is not NULL or empty"""

    def get_invalid(self, session):
        return (
            self.to_check(session)
            .filter((self.column == None) | (self.column == ""))
            .all()
        )

    def description(self):
        return f"{self.column_name} cannot be null or empty for shapes {self.shape_msg}"


class CrossSectionExpectEmptyCheck(CrossSectionBaseCheck):
    """Check if width / height is NULL or empty"""

    def get_invalid(self, session):
        return (
            self.to_check(session)
            .filter((self.column != None) & (self.column != ""))
            .all()
        )

    def description(self):
        return f"{self.column_name} should be null or empty for shapes {self.shape_msg}"


class CrossSectionFloatCheck(CrossSectionBaseCheck):
    """Check that width / height is a valid non-negative float"""

    def get_invalid(self, session):
        invalids = []
        for record in self.to_check(session).filter(
            (self.column != None) & (self.column != "")
        ):
            try:
                value = float(getattr(record, self.column.name))
            except ValueError:
                invalids.append(record)
            else:
                if value < 0:
                    invalids.append(record)

        return invalids

    def description(self):
        return f"{self.column_name} should be a positive number for shapes {self.shape_msg}"


class CrossSectionGreaterZeroCheck(CrossSectionBaseCheck):
    """Check that width / height is larger than 0"""

    def get_invalid(self, session):
        invalids = []
        for record in self.to_check(session).filter(
            (self.column != None) & (self.column != "")
        ):
            try:
                value = float(getattr(record, self.column.name))
            except ValueError:
                continue

            if value <= 0:
                invalids.append(record)
        return invalids

    def description(self):
        return f"{self.column_name} should be greater than zero for shapes {self.shape_msg}"


class CrossSectionFloatListCheck(CrossSectionBaseCheck):
    """Tabulated definitions should use a space for separating the floats."""

    def get_invalid(self, session):
        invalids = []
        for record in self.to_check(session).filter(
            (self.column != None) & (self.column != "")
        ):
            try:
                [float(x) for x in getattr(record, self.column.name).split(" ")]
            except ValueError:
                invalids.append(record)

        return invalids

    def description(self):
        return f"{self.column_name} should contain a space separated list of numbers for shapes {self.shape_msg}"


class CrossSectionEqualElementsCheck(CrossSectionBaseCheck):
    """Tabulated definitions should have equal numbers of width and height elements."""

    def __init__(self, *args, **kwargs):
        super().__init__(column=models.CrossSectionDefinition.width, *args, **kwargs)

    def get_invalid(self, session):
        invalids = []
        for record in self.to_check(session).filter(
            (models.CrossSectionDefinition.width != None)
            & (models.CrossSectionDefinition.width != "")
            & (models.CrossSectionDefinition.height != None)
            & (models.CrossSectionDefinition.height != "")
        ):
            try:
                widths = [float(x) for x in record.width.split(" ")]
                heights = [float(x) for x in record.height.split(" ")]
            except ValueError:
                continue  # other check catches this

            if len(widths) != len(heights):
                invalids.append(record)

        return invalids

    def description(self):
        return f"{self.table.name} width and height should an equal number of elements for shapes {self.shape_msg}"


class CrossSectionIncreasingCheck(CrossSectionBaseCheck):
    """Tabulated definitions should have an increasing list of heights."""

    def get_invalid(self, session):
        invalids = []
        for record in self.to_check(session).filter(
            (self.column != None) & (self.column != "")
        ):
            try:
                values = [
                    float(x) for x in getattr(record, self.column.name).split(" ")
                ]
            except ValueError:
                continue  # other check catches this

            if len(values) > 1 and any(
                x > y for (x, y) in zip(values[:-1], values[1:])
            ):
                invalids.append(record)

        return invalids

    def description(self):
        return f"{self.column_name} should be monotonically increasing for shapes {self.shape_msg}. Maybe the width and height have been interchanged?"


class CrossSectionFirstElementZeroCheck(CrossSectionBaseCheck):
    """Tabulated definitions should start at with 0 height."""

    def get_invalid(self, session):
        invalids = []
        for record in self.to_check(session).filter(
            (self.column != None) & (self.column != "")
        ):
            try:
                values = [
                    float(x) for x in getattr(record, self.column.name).split(" ")
                ]
            except ValueError:
                continue  # other check catches this

            if abs(values[0]) != 0:
                invalids.append(record)

        return invalids

    def description(self):
        return f"The first element of {self.column_name} should equal 0 for shapes {self.shape_msg}. Note that heights are relative to 'reference_level'."


class CrossSectionFirstElementNonZeroCheck(CrossSectionBaseCheck):
    """Tabulated rectangles cannot start with 0 width."""

    def get_invalid(self, session):
        invalids = []
        for record in self.to_check(session).filter(
            (self.column != None) & (self.column != "")
        ):
            try:
                values = [
                    float(x) for x in getattr(record, self.column.name).split(" ")
                ]
            except ValueError:
                continue  # other check catches this

            if abs(values[0]) <= 0:
                invalids.append(record)

        return invalids

    def description(self):
        return f"The first element of {self.column_name} must be larger than 0 for tabulated rectangle shapes. Consider using tabulated trapezium."


class CrossSectionYZHeightCheck(CrossSectionBaseCheck):
    """The height of an yz profile should include 0 and should not have negative
    elements.
    """

    def get_invalid(self, session):
        invalids = []
        for record in self.to_check(session).filter(
            (self.column != None) & (self.column != "")
        ):
            try:
                values = [
                    float(x) for x in getattr(record, self.column.name).split(" ")
                ]
            except ValueError:
                continue  # other check catches this

            if len(values) == 0:
                continue

            if any(x < 0 for x in values) or not any(x == 0 for x in values):
                invalids.append(record)

        return invalids

    def description(self):
        return f"{self.column_name} for YZ profiles should include 0.0 and should not include negative values."


class CrossSectionYZCoordinateCountCheck(CrossSectionBaseCheck):
    """yz profiles should have 3 coordinates (excluding a closing coordinate)"""

    def __init__(self, *args, **kwargs):
        super().__init__(column=models.CrossSectionDefinition.width, *args, **kwargs)

    def get_invalid(self, session):
        invalids = []
        for record in self.to_check(session).filter(
            (models.CrossSectionDefinition.width != None)
            & (models.CrossSectionDefinition.width != "")
            & (models.CrossSectionDefinition.height != None)
            & (models.CrossSectionDefinition.height != "")
        ):
            try:
                widths = [float(x) for x in record.width.split(" ")]
                heights = [float(x) for x in record.height.split(" ")]
            except ValueError:
                continue  # other check catches this

            if len(widths) == 0 or len(widths) != len(heights):
                continue

            is_closed = widths[0] == widths[-1] and heights[0] == heights[-1]
            if len(heights) < (4 if is_closed else 3):
                invalids.append(record)

        return invalids

    def description(self):
        return f"{self.table.name} width and height should contain at least 3 coordinates (excluding closing coordinate) for YZ profiles"


class CrossSectionYZIncreasingWidthIfOpenCheck(CrossSectionBaseCheck):
    """yz profiles should have increasing widths (y) if they are open"""

    def __init__(self, *args, **kwargs):
        super().__init__(column=models.CrossSectionDefinition.width, *args, **kwargs)

    def get_invalid(self, session):
        invalids = []
        for record in self.to_check(session).filter(
            (models.CrossSectionDefinition.width != None)
            & (models.CrossSectionDefinition.width != "")
            & (models.CrossSectionDefinition.height != None)
            & (models.CrossSectionDefinition.height != "")
        ):
            try:
                widths = [float(x) for x in record.width.split(" ")]
                heights = [float(x) for x in record.height.split(" ")]
            except ValueError:
                continue  # other check catches this

            if widths[0] == widths[-1] and heights[0] == heights[-1]:
                continue
            elif len(widths) > 1 and any(
                x >= y for (x, y) in zip(widths[:-1], widths[1:])
            ):
                invalids.append(record)

        return invalids

    def description(self):
        return f"{self.column_name} should be monotonically increasing for open YZ profiles. Perhaps this is actually a closed profile?"



from threedi_modelchecker import exporters
from threedi_modelchecker.checks.base import CheckLevel
from threedi_modelchecker.model_checks import ThreediModelChecker
from threedi_modelchecker.schema import ModelSchema
from threedi_modelchecker.threedi_database import ThreediDatabase

import click


@click.group()
@click.option(
    "-s",
    "--sqlite",
    type=click.Path(exists=True, readable=True),
    help="Path to an sqlite (spatialite) file",
    required=True,
)
@click.pass_context
def threedi_modelchecker(ctx, sqlite):
    """Checks the threedi-model for errors / warnings / info messages"""
    ctx.ensure_object(dict)

    db = ThreediDatabase(sqlite, echo=False)
    ctx.obj["db"] = db


@threedi_modelchecker.command()
@click.option("-f", "--file", help="Write errors to file, instead of stdout")
@click.option(
    "-l",
    "--level",
    type=click.Choice([x.name for x in CheckLevel], case_sensitive=False),
    default="ERROR",
    help="Minimum check level.",
)
@click.pass_context
def check(ctx, file, level):
    """Checks the threedi model schematisation for errors."""
    level = level.upper()
    if level == "ERROR":
        msg = "errors"
    elif level == "WARNING":
        msg = "errors or warnings"
    else:
        msg = "errors, warnings or info messages"
    click.echo("Parsing schematisation for any %s" % msg)
    if file:
        click.echo("Model errors will be written to %s" % file)

    mc = ThreediModelChecker(ctx.obj["db"])
    model_errors = mc.errors(level=level)

    if file:
        exporters.export_to_file(model_errors, file)
    else:
        exporters.print_errors(model_errors)

    click.echo("Finished processing model")


@threedi_modelchecker.command()
@click.option(
    "-r", "--revision", default="head", help="The schema revision to migrate to"
)
@click.option("--backup/--no-backup", default=True)
@click.option("--set-views/--no-set-views", default=True)
@click.option(
    "--upgrade-spatialite-version/--no-upgrade-spatialite-version", default=False
)
@click.pass_context
def migrate(ctx, revision, backup, set_views, upgrade_spatialite_version):
    """Migrate the threedi model schematisation to the latest version."""
    schema = ModelSchema(ctx.obj["db"])
    click.echo("The current schema revision is: %s" % schema.get_version())
    click.echo("Running alembic upgrade script...")
    schema.upgrade(
        revision=revision,
        backup=backup,
        set_views=set_views,
        upgrade_spatialite_version=upgrade_spatialite_version,
    )
    click.echo("The migrated schema revision is: %s" % schema.get_version())


@threedi_modelchecker.command()
@click.pass_context
def index(ctx):
    """Set the indexes of a threedi model schematisation."""
    schema = ModelSchema(ctx.obj["db"])
    click.echo("Recovering indexes...")
    schema.set_spatial_indexes()
    click.echo("Done.")


if __name__ == "__main__":
    threedi_modelchecker()



from sqlalchemy.orm import Query
from sqlalchemy.orm.session import Session
from threedi_api_client.openapi.models.lateral import Lateral
from threedi_modelchecker.simulation_templates.exceptions import (
    SchematisationError,
)
from threedi_modelchecker.simulation_templates.utils import parse_timeseries
from threedi_modelchecker.simulation_templates.utils import (
    strip_dict_none_values,
)
from threedi_modelchecker.threedi_model.models import Lateral1d
from threedi_modelchecker.threedi_model.models import Lateral2D
from typing import List

import json


def lateral_1d_to_api_lateral(lateral_1d: Lateral1d) -> Lateral:
    try:
        values = parse_timeseries(lateral_1d.timeseries)
    except (ValueError, TypeError):
        raise SchematisationError(
            f"Incorrect timeseries format for lateral 1D with id: {lateral_1d.id}"
        )

    offset = values[0][0]

    if offset > 0:
        # Shift timeseries to start at t=0
        values = [[x[0] - offset, x[1]] for x in values]

    return Lateral(
        connection_node=int(lateral_1d.connection_node_id),
        offset=int(values[0][0]),
        values=values,
        units="m3/s",
        interpolate=False,
    )


def lateral_2d_to_api_lateral(lateral_2d: Lateral2D, session: Session) -> Lateral:
    try:
        values = parse_timeseries(lateral_2d.timeseries)
    except (ValueError, TypeError):
        raise SchematisationError(
            f"Incorrect timeseries format for lateral 2D with id: {lateral_2d.id}"
        )

    offset = values[0][0]

    if offset > 0:
        # Shift timeseries to start at t=0
        values = [[x[0] - offset, x[1]] for x in values]

    # x,y is correct (4.294348493375471, 52.033176579129936) until we alter the API
    point = json.loads(session.scalar(lateral_2d.the_geom.ST_AsGeoJSON()))

    return Lateral(
        offset=int(values[0][0]),
        values=values,
        point=point,
        units="m3/s",
        interpolate=False,
    )


class LateralsExtractor(object):
    def __init__(self, session: Session):
        self.session = session
        self._laterals_2d = None
        self._laterals_1d = None

    @property
    def laterals_2d(self) -> List[Lateral]:
        if self._laterals_2d is None:
            laterals_2d = Query(Lateral2D).with_session(self.session).all()
            self._laterals_2d = [
                lateral_2d_to_api_lateral(x, self.session) for x in laterals_2d
            ]

        return self._laterals_2d

    @property
    def laterals_1d(self) -> List[Lateral]:
        if self._laterals_1d is None:
            laterals_1d = Query(Lateral1d).with_session(self.session).all()
            self._laterals_1d = [lateral_1d_to_api_lateral(x) for x in laterals_1d]

        return self._laterals_1d

    def all_laterals(self) -> List[Lateral]:
        return self.laterals_2d + self.laterals_1d

    def as_list(self) -> List[dict]:
        json_laterals = []
        for lateral in self.all_laterals():
            json_lateral = lateral.to_dict()
            strip_dict_none_values(json_lateral)
            json_laterals.append(json_lateral)
        return json_laterals



from .base import BaseCheck
from dataclasses import dataclass
from pathlib import Path
from threedi_modelchecker.interfaces import GDALRasterInterface
from threedi_modelchecker.interfaces import RasterInterface
from typing import Optional
from typing import Set
from typing import Type


class Context:
    pass


@dataclass
class ServerContext(Context):
    available_rasters: Set[str]
    raster_interface: Type[RasterInterface] = GDALRasterInterface


@dataclass
class LocalContext(Context):
    base_path: Path
    raster_interface: Type[RasterInterface] = GDALRasterInterface


class BaseRasterCheck(BaseCheck):
    """Baseclass for all raster checks.

    Because these checks are different on local and server systems, subclasses may
    implement 2 methods: is_valid_local and/or is_valid_local.
    """

    def to_check(self, session):
        return (
            super()
            .to_check(session)
            .filter(
                self.column != None,
                self.column != "",
            )
        )

    def get_invalid(self, session):
        context = session.model_checker_context
        raster_interface = context.raster_interface
        if not raster_interface.available():
            return []
        if isinstance(context, ServerContext):
            # max 1 record; we can only have 1 raster per type on the server
            records = list(self.to_check(session).limit(1).all())
            paths = [self.get_path_server(x, context) for x in records]
        else:
            records = list(self.to_check(session).all())
            paths = [self.get_path_local(x, context) for x in records]
        return [
            record
            for (record, path) in zip(records, paths)
            if path is not None and not self.is_valid(path, raster_interface)
        ]

    def get_path_local(self, record, context: LocalContext) -> Optional[str]:
        abs_path = context.base_path / getattr(record, self.column.name)
        if abs_path.exists():
            return str(abs_path)

    def get_path_server(self, record, context: ServerContext) -> str:
        if isinstance(context.available_rasters, dict):
            return context.available_rasters.get(self.column.name)

    def is_valid(self, path: str, interface_cls: Type[RasterInterface]):
        return True


class RasterExistsCheck(BaseRasterCheck):
    """Check whether a file referenced in given Column exists.

    In order to perform this check, the SQLAlchemy session requires a
    `model_checker_context` attribute, which is set automatically by the
    ThreediModelChecker and contains either a LocalContext or
    ServerContextinstance.
    """

    def get_invalid(self, session):
        context = session.model_checker_context
        if isinstance(context, ServerContext):
            if self.column.name in context.available_rasters:
                return []
            else:
                return self.to_check(session).all()
        else:
            records = list(self.to_check(session).all())
            paths = [self.get_path_local(x, context) for x in records]
            return [record for (record, path) in zip(records, paths) if path is None]

    def description(self):
        return f"The file in {self.column_name} is not present"


class RasterIsValidCheck(BaseRasterCheck):
    """Check whether a file is a geotiff.

    Only works locally.
    """

    def is_valid(self, path: str, interface_cls: Type[RasterInterface]):
        with interface_cls(path) as raster:
            return raster.is_valid_geotiff

    def description(self):
        return f"The file in {self.column_name} is not a valid GeoTIFF file"


class RasterHasOneBandCheck(BaseRasterCheck):
    """Check whether a raster has a single band.

    Only works locally.
    """

    def is_valid(self, path: str, interface_cls: Type[RasterInterface]):
        with interface_cls(path) as raster:
            if not raster.is_valid_geotiff:
                return True
            return raster.band_count == 1

    def description(self):
        return f"The file in {self.column_name} has multiple or no bands."


class RasterHasProjectionCheck(BaseRasterCheck):
    """Check whether a raster has a projected coordinate system.

    Only works locally.
    """

    def is_valid(self, path: str, interface_cls: Type[RasterInterface]):
        with interface_cls(path) as raster:
            if not raster.is_valid_geotiff:
                return True
            return raster.is_geographic is not None

    def description(self):
        return f"The file in {self.column_name} has no CRS."


class RasterIsProjectedCheck(BaseRasterCheck):
    """Check whether a raster has a projected coordinate system.

    Only works locally.
    """

    def is_valid(self, path: str, interface_cls: Type[RasterInterface]):
        with interface_cls(path) as raster:
            if not raster.is_valid_geotiff:
                return True
            return raster.is_geographic is not True

    def description(self):
        return f"The file in {self.column_name} does not use a projected CRS."


class RasterSquareCellsCheck(BaseRasterCheck):
    """Check whether a raster has square cells (pixels)

    Only works locally.
    """

    def __init__(self, *args, decimals=7, **kwargs):
        super().__init__(*args, **kwargs)
        self.decimals = decimals

    def is_valid(self, path: str, interface_cls: Type[RasterInterface]):
        with interface_cls(path) as raster:
            if not raster.is_valid_geotiff:
                return True
            dx, dy = raster.pixel_size
            return dx is not None and round(dx, self.decimals) == round(
                dy, self.decimals
            )

    def description(self):
        return f"The raster in {self.column_name} has non-square raster cells."


class RasterRangeCheck(BaseRasterCheck):
    """Check whether a raster has values outside of provided range.

    Also fails when there are no values in the raster. Only works locally.
    """

    def __init__(
        self,
        min_value=None,
        max_value=None,
        left_inclusive=True,
        right_inclusive=True,
        message=None,
        *args,
        **kwargs,
    ):
        if min_value is None and max_value is None:
            raise ValueError("Please supply at least one of {min_value, max_value}.")
        self.min_value = min_value
        self.max_value = max_value
        self.left_inclusive = left_inclusive
        self.right_inclusive = right_inclusive
        self.message = message
        super().__init__(*args, **kwargs)

    def is_valid(self, path: str, interface_cls: Type[RasterInterface]):
        with interface_cls(path) as raster:
            if not raster.is_valid_geotiff:
                return True
            try:
                raster_min, raster_max = raster.min_max
            except RasterInterface.NoData:
                return False  # no data in the raster is invalid too

        if raster_min is None or raster_max is None:
            return False
        if self.min_value is not None:
            if self.left_inclusive and raster_min < self.min_value:
                return False
            if not self.left_inclusive and raster_min <= self.min_value:
                return False
        if self.max_value is not None:
            if self.right_inclusive and raster_max > self.max_value:
                return False
            if not self.right_inclusive and raster_max >= self.max_value:
                return False

        return True

    def description(self):
        if self.message:
            return self.message
        parts = []
        if self.min_value is not None:
            parts.append(f"{'<' if self.left_inclusive else '<='}{self.min_value}")
        if self.max_value is not None:
            parts.append(f"{'>' if self.right_inclusive else '>='}{self.max_value}")
        return f"{self.column_name} has values {' and/or '.join(parts)} or is empty"


@dataclass
class GDALUnavailable:
    id: int


class GDALAvailableCheck(BaseRasterCheck):
    """Checks whether GDAL is available"""

    def get_invalid(self, session):
        context = session.model_checker_context
        raster_interface = context.raster_interface
        if not raster_interface.available():
            return [GDALUnavailable(id=1)]
        else:
            return []

    def description(self) -> str:
        return "raster checks require GDAL to be available"



"""clean unused connected points and transform remaining to potential breaches

Revision ID: 0213
Revises: 0212
Create Date: 2022-12-21 14:54:00

"""
from alembic import op
from geoalchemy2.types import Geometry
from sqlalchemy import Column
from sqlalchemy import Float
from sqlalchemy import ForeignKey
from sqlalchemy import func
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session

import logging


# revision identifiers, used by Alembic.
revision = "0213"
down_revision = "0212"
branch_labels = None
depends_on = None

logger = logging.getLogger(__name__)

## Copy of the ORM at this point in time:

Base = declarative_base()


class Levee(Base):
    __tablename__ = "v2_levee"
    id = Column(Integer, primary_key=True)
    code = Column(String(100))
    crest_level = Column(Float)
    the_geom = Column(
        Geometry(
            geometry_type="LINESTRING", srid=4326, spatial_index=True, management=True
        ),
        nullable=False,
    )
    material = Column(Integer)
    max_breach_depth = Column(Float)


class CalculationPoint(Base):
    __tablename__ = "v2_calculation_point"
    id = Column(Integer, primary_key=True)

    content_type_id = Column(Integer)
    user_ref = Column(String(80), nullable=False)
    calc_type = Column(Integer)
    the_geom = Column(
        Geometry(geometry_type="POINT", srid=4326),
        nullable=False,
    )


class ConnectedPoint(Base):
    __tablename__ = "v2_connected_pnt"
    id = Column(Integer, primary_key=True)

    calculation_pnt_id = Column(
        Integer, ForeignKey(CalculationPoint.__tablename__ + ".id"), nullable=False
    )
    levee_id = Column(Integer, ForeignKey(Levee.__tablename__ + ".id"))

    exchange_level = Column(Float)
    the_geom = Column(
        Geometry(geometry_type="POINT", srid=4326),
        nullable=False,
    )


class PotentialBreach(Base):
    __tablename__ = "v2_potential_breach"
    id = Column(Integer, primary_key=True)
    code = Column(String(100))
    display_name = Column(String(255))
    exchange_level = Column(Float)
    maximum_breach_depth = Column(Float)
    levee_material = Column(Integer)
    the_geom = Column(
        Geometry(geometry_type="LINESTRING", srid=4326, from_text="ST_GeomFromEWKB"),
        nullable=False,
    )
    channel_id = Column(Integer, nullable=False)


class ConnectionNode(Base):
    __tablename__ = "v2_connection_nodes"
    id = Column(Integer, primary_key=True)
    the_geom = Column(
        Geometry(geometry_type="POINT", srid=4326, spatial_index=True, management=True),
        nullable=False,
    )


class Manhole(Base):
    __tablename__ = "v2_manhole"

    id = Column(Integer, primary_key=True)
    connection_node_id = Column(
        Integer,
        ForeignKey(ConnectionNode.__tablename__ + ".id"),
        nullable=False,
        unique=True,
    )


class Channel(Base):
    __tablename__ = "v2_channel"

    id = Column(Integer, primary_key=True)
    calculation_type = Column(Integer, nullable=False)
    dist_calc_points = Column(Float, nullable=True)
    connection_node_start_id = Column(Integer, nullable=False)
    connection_node_end_id = Column(Integer, nullable=False)
    the_geom = Column(
        Geometry(
            geometry_type="LINESTRING", srid=4326, spatial_index=True, management=True
        ),
        nullable=True,
    )


def parse_connected_point_user_ref(user_ref: str):
    """Return content_type, content_id, node_number from a user_ref.

    Raises Exception for various parse errors.

    Example
    -------
    >>> parse_connected_point_user_ref("201#123#v2_channels#4)
    ContentType.TYPE_V2_CHANNEL, 123, 3
    """
    _, id_str, type_str, node_nr = user_ref.split("#")
    return type_str, int(id_str), int(node_nr)


def clean_connected_points(session):
    conn_point_ids = [
        x[0]
        for x in session.query(ConnectedPoint.id)
        .join(CalculationPoint, isouter=True)
        .filter(
            (ConnectedPoint.the_geom != CalculationPoint.the_geom)
            | (ConnectedPoint.exchange_level > -9999.0)
            | (ConnectedPoint.levee_id != None)
        )
        .all()
    ]
    session.query(ConnectedPoint).filter(
        ConnectedPoint.id.notin_(conn_point_ids)
    ).delete(synchronize_session="fetch")
    calc_point_ids = [
        x[0]
        for x in session.query(ConnectedPoint.calculation_pnt_id)
        .filter(ConnectedPoint.id.in_(conn_point_ids))
        .all()
    ]
    session.query(CalculationPoint).filter(
        CalculationPoint.id.notin_(calc_point_ids)
    ).delete(synchronize_session="fetch")
    return conn_point_ids


def get_channel_id(session, user_ref):
    """Get channel id and index into channel nodes"""
    type_ref, pk, node_nr = parse_connected_point_user_ref(user_ref)
    if type_ref == "v2_channel":
        return pk, node_nr - 1
    elif type_ref == "v2_manhole":
        return get_channel_id_manhole(session, pk)
    return None, None


def get_channel_id_manhole(session, pk):
    obj = session.query(Manhole).filter(Manhole.id == pk).first()
    if obj is None:
        return None, None
    connection_node_id = obj.connection_node_id

    # find a channel connected to this connection node, with connected calc type
    channels = (
        session.query(Channel)
        .filter(
            (
                (Channel.connection_node_start_id == connection_node_id)
                | (Channel.connection_node_end_id == connection_node_id)
            )
            & Channel.calculation_type.in_([102, 105])
        )
        .all()
    )
    if len(channels) == 0:
        return None, None

    # prefer double connected, and then prefer lowest id
    channel = sorted(channels, key=lambda x: (-x.calculation_type, x.id))[0]
    if channel.connection_node_start_id == connection_node_id:
        node_idx = 0
    else:
        node_idx = -1
    return channel.id, node_idx


def scalar_subquery(query):
    # compatibility between sqlalchemy 1.3 and 1.4
    try:
        return query.scalar_subquery()
    except AttributeError:
        return query.as_scalar()


def get_breach_line_geom(session, conn_point_id, channel_id, node_idx):
    channel = scalar_subquery(
        session.query(Channel.the_geom).filter(Channel.id == channel_id)
    ).limit(1)
    if node_idx == 0:
        breach_line_start = func.ST_PointN(channel, 1)
    elif node_idx == -1:
        breach_line_start = func.ST_PointN(channel, func.ST_NPoints(channel))
    else:
        breach_line_start = func.Snap(CalculationPoint.the_geom, channel, 1e-7)

    return (
        session.query(
            func.AsEWKB(func.MakeLine(breach_line_start, ConnectedPoint.the_geom))
        )
        .join(CalculationPoint)
        .filter(ConnectedPoint.id == conn_point_id)
        .one()[0]
    )


def to_potential_breach(session, conn_point_id):
    connected_point, calculation_point, levee = (
        session.query(
            ConnectedPoint,
            CalculationPoint,
            Levee,
        )
        .join(CalculationPoint)
        .join(Levee, isouter=True)
        .filter(ConnectedPoint.id == conn_point_id)
        .one()
    )

    channel_id, node_idx = get_channel_id(session, calculation_point.user_ref)
    if channel_id is None:
        return

    line_geom = get_breach_line_geom(session, conn_point_id, channel_id, node_idx)

    if connected_point.exchange_level not in (None, -9999.0):
        exchange_level = connected_point.exchange_level
    elif levee is not None:
        exchange_level = levee.crest_level
    else:
        exchange_level = None

    if levee is not None:
        maximum_breach_depth = levee.max_breach_depth
    else:
        maximum_breach_depth = None

    if exchange_level == -9999.0:
        exchange_level = None
    if maximum_breach_depth == -9999.0:
        maximum_breach_depth = None

    return PotentialBreach(
        code="#".join([str(connected_point.id), calculation_point.user_ref])[:100],
        exchange_level=exchange_level,
        maximum_breach_depth=maximum_breach_depth,
        levee_material=levee.material if levee is not None else None,
        the_geom=line_geom,
        channel_id=channel_id,
    )


def upgrade():
    session = Session(bind=op.get_bind())

    conn_point_ids = clean_connected_points(session)
    for conn_point_id in conn_point_ids:
        breach = to_potential_breach(session, conn_point_id)
        if breach is None:
            logger.warning(
                "Connected Point %d will be removed because it "
                "cannot be related to a channel. This may influence the "
                "1D-2D exchange of the model.",
                conn_point_id,
            )
        else:
            session.add(breach)
    session.flush()


def downgrade():
    pass



from .exceptions import SchematisationError
from pathlib import Path
from sqlalchemy.orm import Query
from sqlalchemy.orm.session import Session
from threedi_modelchecker.simulation_templates.boundaries.extractor import (
    BoundariesExtractor,
)
from threedi_modelchecker.simulation_templates.initial_waterlevels.extractor import (
    InitialWaterlevelExtractor,
)
from threedi_modelchecker.simulation_templates.laterals.dwf_calculator import (
    DWFCalculator,
)
from threedi_modelchecker.simulation_templates.laterals.extractor import (
    LateralsExtractor,
)
from threedi_modelchecker.simulation_templates.models import Events
from threedi_modelchecker.simulation_templates.models import (
    GlobalSettingOption,
)
from threedi_modelchecker.simulation_templates.models import SimulationTemplate
from threedi_modelchecker.simulation_templates.settings.extractor import (
    SettingsExtractor,
)
from threedi_modelchecker.simulation_templates.structure_controls.extractor import (
    StructureControlExtractor,
)
from threedi_modelchecker.threedi_database import ThreediDatabase
from threedi_modelchecker.threedi_model.models import GlobalSetting
from typing import List
from typing import Optional


__all__ = ["SimulationTemplateExtractor"]


class SimulationTemplateExtractor(object):
    def __init__(self, sqlite_path: Path):
        """
        param global_settings_id: if None the first global setting entry is taken (default)
        """
        self.sqlite_path = sqlite_path
        self.database = ThreediDatabase(str(self.sqlite_path))

    def _extract_simulation_template(
        self, session: Session, global_settings_id: Optional[int] = None
    ) -> SimulationTemplate:
        """
        Extract a SimulationTemplate instance using the given database session
        """
        qr = Query(GlobalSetting).with_session(session)
        if global_settings_id is not None:
            qr = qr.filter(GlobalSetting.id == global_settings_id)
        global_settings: GlobalSetting = qr.first()

        if global_settings is None:
            raise SchematisationError(
                f"Global settings with id: {global_settings_id} not found."
            )

        dwf_laterals = DWFCalculator(session, global_settings.use_0d_inflow).laterals
        initial_waterlevels = InitialWaterlevelExtractor(session, global_settings_id)
        settings = SettingsExtractor(session, global_settings.id)

        return SimulationTemplate(
            events=Events(
                structure_controls=StructureControlExtractor(
                    session, control_group_id=global_settings.control_group_id
                ).all_controls(),
                laterals=LateralsExtractor(session).as_list(),
                dwf_laterals=dwf_laterals,
                boundaries=BoundariesExtractor(session).as_list(),
            ),
            settings=settings.all_settings(),
            initial_waterlevels=initial_waterlevels.all_initial_waterlevels(),
        )

    def _get_global_settings_options(
        self, session: Session
    ) -> List[GlobalSettingOption]:
        return [
            GlobalSettingOption(x.id, x.name)
            for x in Query(GlobalSetting).with_session(session)
        ]

    def global_settings_options(self) -> List[GlobalSettingOption]:
        try:
            session = self.database.get_session()
            return self._get_global_settings_options(session)
        finally:
            session.close()

    def extract(self, global_settings_id: Optional[int] = None) -> SimulationTemplate:
        """
        Return simulation template for
        """
        try:
            session = self.database.get_session()
            return self._extract_simulation_template(session, global_settings_id)
        finally:
            session.close()



from sqlalchemy.orm import Session
from threedi_modelchecker.threedi_model.constants import InflowType
from typing import Dict
from typing import List
from typing import Optional
from typing import Union


INFLOW_TABLE_NAME_BASES = {
    InflowType.IMPERVIOUS_SURFACE: "impervious_surface",
    InflowType.SURFACE: "surface",
}


# Default values
DWF_FACTORS = [
    [0, 0.03],
    [1, 0.015],
    [2, 0.01],
    [3, 0.01],
    [4, 0.005],
    [5, 0.005],
    [6, 0.025],
    [7, 0.080],
    [8, 0.075],
    [9, 0.06],
    [10, 0.055],
    [11, 0.05],
    [12, 0.045],
    [13, 0.04],
    [14, 0.04],
    [15, 0.035],
    [16, 0.035],
    [17, 0.04],
    [18, 0.055],
    [19, 0.08],
    [20, 0.07],
    [21, 0.055],
    [22, 0.045],
    [23, 0.04],
    [24, 0.0],  # Timeseries for laterals should contain 25 values
]


def read_dwf_per_node(
    session: Session, inflow_type: InflowType
) -> List[Union[int, float]]:
    """Obtains the total dry weather flow in m3/d per connection node from a 3Di model sqlite-file."""

    basename = INFLOW_TABLE_NAME_BASES[inflow_type]
    query = f"""
        SELECT 	map.connection_node_id,
                sum(surf.dry_weather_flow * surf.nr_of_inhabitants * map.percentage/100)/1000 AS dwf
        FROM 	v2_{basename} AS surf
        JOIN 	v2_{basename}_map AS map
        ON 		surf.id = map.{basename}_id
        WHERE 	surf.dry_weather_flow IS NOT NULL
                and surf.nr_of_inhabitants != 0
                and surf.nr_of_inhabitants IS NOT NULL
                and map.percentage IS NOT NULL
        GROUP BY map.connection_node_id
    """
    dwf_per_node = [
        [connection_node_id, weighted_flow_sum]
        for connection_node_id, weighted_flow_sum in session.execute(query)
    ]

    return dwf_per_node


def generate_dwf_laterals(session: Session, inflow_type: InflowType) -> List[Dict]:
    """Generate dry weather flow laterals from spatialite"""
    dwf_per_node = read_dwf_per_node(session, inflow_type)
    dwf_laterals = []

    # Generate lateral for each connection node
    for node_id, flow in dwf_per_node:
        values = [[hour * 3600, flow * factor / 3600] for hour, factor in DWF_FACTORS]
        dwf_laterals.append(
            dict(
                offset=0,
                values=values,
                units="m3/s",
                connection_node=node_id,
                interpolate=False,
            )
        )

    return dwf_laterals


class DWFCalculator:
    """Calculate dry weather flow (DWF) from sqlite."""

    def __init__(self, session: Session, inflow_type: InflowType) -> None:
        self.session = session
        self.inflow_type = inflow_type
        self._laterals = None

    @property
    def laterals(self) -> List[Optional[Dict]]:
        if self.inflow_type not in [InflowType.SURFACE, InflowType.IMPERVIOUS_SURFACE]:
            return []

        if self._laterals is None:
            self._laterals = generate_dwf_laterals(self.session, self.inflow_type)

        return self._laterals



from .base import BaseCheck


def parse_timeseries(timeseries_str):
    if not timeseries_str:
        return []
    output = []
    for line in timeseries_str.split():
        timestep, value = line.split(",")
        timestep = int(timestep.strip())
        output.append([timestep, float(value.strip())])
    return output


class TimeseriesRowCheck(BaseCheck):
    """Check that each record in a timeserie contains 2 elements"""

    def get_invalid(self, session):
        invalid_timeseries = []

        for row in self.to_check(session).all():
            timeserie = row.timeseries

            if not timeserie:
                continue

            if any(len(x.split(",")) != 2 for x in timeserie.split()):
                invalid_timeseries.append(row)

        return invalid_timeseries

    def description(self):
        return (
            f"{self.column_name} must contain 2 elements per line separated by a comma"
        )


class TimeseriesTimestepCheck(BaseCheck):
    """Check that each record in a timeserie starts with an integer >= 0"""

    def get_invalid(self, session):
        invalid_timeseries = []

        for row in self.to_check(session).all():
            timeserie = row.timeseries

            if not timeserie:
                continue

            for timeseries_row in timeserie.split():
                elems = timeseries_row.split(",")
                if len(elems) != 2:
                    continue  # checked elsewhere

                try:
                    timestep = int(elems[0].strip())
                except ValueError:
                    invalid_timeseries.append(row)
                    continue

                if timestep < 0:
                    invalid_timeseries.append(row)

        return invalid_timeseries

    def description(self):
        return (
            f"{self.column_name} contains an invalid timestep, expected an integer >= 0"
        )


class TimeseriesValueCheck(BaseCheck):
    """Check that each record in a timeserie ends with a float"""

    def get_invalid(self, session):
        invalid_timeseries = []

        for row in self.to_check(session).all():
            timeserie = row.timeseries

            if not timeserie:
                continue

            for timeseries_row in timeserie.split():
                elems = timeseries_row.split(",")
                if len(elems) != 2:
                    continue  # checked elsewhere

                try:
                    value = float(elems[1].strip())
                except ValueError:
                    invalid_timeseries.append(row)
                    continue

                if str(value) in {"nan", "inf", "-inf"}:
                    invalid_timeseries.append(row)

        return invalid_timeseries

    def description(self):
        return f"{self.column_name} contains an invalid value, expected a float"


class TimeseriesIncreasingCheck(BaseCheck):
    """The timesteps in a timeseries should increase"""

    def get_invalid(self, session):
        invalid_timeseries = []

        for row in self.to_check(session).all():
            timeserie = row.timeseries
            try:
                timesteps = [x[0] for x in parse_timeseries(timeserie)]
            except (ValueError, TypeError):
                continue  # other checks will catch these

            if len(timesteps) < 2:
                continue

            if not all(b > a for (a, b) in zip(timesteps[:-1], timesteps[1:])):
                invalid_timeseries.append(row)

        return invalid_timeseries

    def description(self):
        return f"{self.column_name} should be monotonically increasing"


class TimeseriesStartsAtZeroCheck(BaseCheck):
    """The timesteps in a timeseries should start at 0"""

    def get_invalid(self, session):
        invalid_timeseries = []

        for row in self.to_check(session).all():
            timeserie = row.timeseries
            try:
                timesteps = [x[0] for x in parse_timeseries(timeserie)]
            except (ValueError, TypeError):
                continue  # other checks will catch these

            if len(timesteps) == 0:
                continue

            if timesteps[0] != 0:
                invalid_timeseries.append(row)

        return invalid_timeseries

    def description(self):
        return f"{self.column_name} should be start at timestamp 0"



from typing import Any
from typing import Dict
from typing import List


def parse_timeseries(
    timeseries_str, time_input_unit="minute", time_output_unit="second"
) -> List[List[float]]:
    """Create a list of 2-list [timestep (seconds), value (mm/hour)]."""
    if not timeseries_str:
        return [[]]
    output = []
    for line in timeseries_str.split():
        timestep, value = line.split(",")
        if time_input_unit == "second":
            timestep = int(timestep.strip())
        else:
            timestep = int(timestep.strip())
            if time_output_unit == "second":
                timestep *= 60
        output.append([timestep, float(value.strip())])
    return output


def strip_dict_none_values(value: Any):
    if isinstance(value, List):
        for x in value:
            strip_dict_none_values(x)
    if isinstance(value, Dict):
        to_delete = []
        for k, v in value.items():
            if v == None:
                to_delete.append(k)
            else:
                strip_dict_none_values(v)

        for x in to_delete:
            del value[x]



from .errors import MigrationMissingError
from .spatialite_versions import copy_models
from .spatialite_versions import get_spatialite_version
from .threedi_model import constants
from .threedi_model import models
from .threedi_model import views
from alembic import command as alembic_command
from alembic.config import Config
from alembic.environment import EnvironmentContext
from alembic.migration import MigrationContext
from alembic.script import ScriptDirectory
from geoalchemy2.types import Geometry
from sqlalchemy import Column
from sqlalchemy import func
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import Table

import warnings


__all__ = ["ModelSchema"]


def get_alembic_config(connection=None):
    alembic_cfg = Config()
    alembic_cfg.set_main_option("script_location", "threedi_modelchecker:migrations")
    alembic_cfg.set_main_option("version_table", constants.VERSION_TABLE_NAME)
    if connection is not None:
        alembic_cfg.attributes["connection"] = connection
    return alembic_cfg


def get_schema_version():
    """Returns the version of the schema in this library"""
    config = get_alembic_config()
    script = ScriptDirectory.from_config(config)
    with EnvironmentContext(config=config, script=script) as env:
        return int(env.get_head_revision())


def _upgrade_database(db, revision="head", unsafe=True):
    """Upgrade ThreediDatabase instance"""
    engine = db.get_engine()

    with engine.begin() as connection:
        if unsafe:
            # Speed up by journalling in memory; in case of a crash the database
            # will likely go corrupt.
            # NB: This setting is scoped to this connection.
            connection.execute("PRAGMA journal_mode = MEMORY")
        config = get_alembic_config(connection)
        alembic_command.upgrade(config, revision)


def _recreate_views(db, file_version):
    """Recreate predefined views in a ThreediDatabase instance"""
    engine = db.get_engine()

    with engine.begin() as connection:
        for (name, view) in views.ALL_VIEWS.items():
            connection.execute(f"DROP VIEW IF EXISTS {name}")
            connection.execute(
                f"DELETE FROM views_geometry_columns WHERE view_name = '{name}'"
            )
            connection.execute(f"CREATE VIEW {name} AS {view['definition']}")
            if file_version == 3:
                connection.execute(
                    f"INSERT INTO views_geometry_columns (view_name, view_geometry,view_rowid,f_table_name,f_geometry_column) VALUES('{name}', '{view['view_geometry']}', '{view['view_rowid']}', '{view['f_table_name']}', '{view['f_geometry_column']}')"
                )
            else:
                connection.execute(
                    f"INSERT INTO views_geometry_columns (view_name, view_geometry,view_rowid,f_table_name,f_geometry_column,read_only) VALUES('{name}', '{view['view_geometry']}', '{view['view_rowid']}', '{view['f_table_name']}', '{view['f_geometry_column']}', 0)"
                )
        for name in views.VIEWS_TO_DELETE:
            connection.execute(f"DROP VIEW IF EXISTS {name}")
            connection.execute(
                f"DELETE FROM views_geometry_columns WHERE view_name = '{name}'"
            )


def _ensure_spatial_index(connection, column):
    """Ensure presence of a spatial index for given geometry olumn"""
    if (
        connection.execute(
            func.RecoverSpatialIndex(column.table.name, column.name)
        ).scalar()
        is not None
    ):
        return False

    idx_name = f"{column.table.name}_{column.name}"
    connection.execute(f"DROP TABLE IF EXISTS idx_{idx_name}")
    for prefix in {"gii_", "giu_", "gid_"}:
        connection.execute(f"DROP TRIGGER IF EXISTS {prefix}{idx_name}")
    if (
        connection.execute(
            func.CreateSpatialIndex(column.table.name, column.name)
        ).scalar()
        != 1
    ):
        raise RuntimeError(f"Spatial index creation for {idx_name} failed")

    return True


def _ensure_spatial_indexes(db):
    """Ensure presence of spatial indexes for all geometry columns"""
    created = False
    engine = db.get_engine()

    with engine.begin() as connection:
        for model in models.DECLARED_MODELS:
            geom_columns = [
                x for x in model.__table__.columns if type(x.type) == Geometry
            ]
            if len(geom_columns) > 1:
                # Pragmatic fix: spatialindex breaks on multiple geometry columns per table
                geom_columns = [x for x in geom_columns if x.name == "the_geom"]
            if geom_columns:
                created &= _ensure_spatial_index(connection, geom_columns[0])

        if created:
            connection.execute("VACUUM")


class ModelSchema:
    def __init__(self, threedi_db, declared_models=models.DECLARED_MODELS):
        self.db = threedi_db
        self.declared_models = declared_models

    def _get_version_old(self):
        """The version of the database using the old 'south' versioning."""
        south_migrationhistory = Table(
            "south_migrationhistory", MetaData(), Column("id", Integer)
        )
        engine = self.db.get_engine()
        if not self.db.has_table("south_migrationhistory"):
            return
        with engine.connect() as connection:
            query = south_migrationhistory.select().order_by(
                south_migrationhistory.columns["id"].desc()
            )
            versions = list(connection.execute(query.limit(1)))
            if len(versions) == 1:
                return versions[0][0]
            else:
                return None

    def get_version(self):
        """Returns the id (integer) of the latest migration"""
        with self.db.get_engine().connect() as connection:
            context = MigrationContext.configure(
                connection, opts={"version_table": constants.VERSION_TABLE_NAME}
            )
            version = context.get_current_revision()

        if version is not None:
            return int(version)
        else:
            return self._get_version_old()

    def upgrade(
        self,
        revision="head",
        backup=True,
        set_views=True,
        upgrade_spatialite_version=False,
    ):
        """Upgrade the database to the latest version.

        This requires either a completely empty database or a database with its
        current schema version at least 174 (the latest migration of the old
        model databank).

        The upgrade is done using database transactions. However, for SQLite,
        database transactions are only partially supported. To ensure that the
        database file does not become corrupt, enable the "backup" parameter.
        If the database is temporary already (or if it is PostGIS), disable
        it.

        Specify 'set_views=True' to also (re)create views after the upgrade.
        This is not compatible when upgrading to a different version than the
        latest version.

        Specify 'upgrade_spatialite_version=True' to also upgrade the
        spatialite file version after the upgrade.
        """
        if upgrade_spatialite_version and not set_views:
            raise ValueError(
                "Cannot upgrade the spatialite version without setting the views."
            )
        v = self.get_version()
        if v is not None and v < constants.LATEST_SOUTH_MIGRATION_ID:
            raise MigrationMissingError(
                f"The modelchecker cannot update versions below "
                f"{constants.LATEST_SOUTH_MIGRATION_ID}. Please consult the "
                f"3Di documentation on how to update legacy databases."
            )
        if set_views and revision not in ("head", get_schema_version()):
            raise ValueError(f"Cannot set views when upgrading to version '{revision}'")
        if backup:
            with self.db.file_transaction() as work_db:
                _upgrade_database(work_db, revision=revision, unsafe=True)
        else:
            _upgrade_database(self.db, revision=revision, unsafe=False)

        if upgrade_spatialite_version:
            self.upgrade_spatialite_version()
        elif set_views:
            self.set_views()

    def validate_schema(self):
        """Very basic validation of 3Di schema.

        Check that the database has the latest migration applied. If the
        latest migrations is applied, we assume the database also contains all
        tables and columns defined in threedi_model.models.py.

        :return: True if the threedi_db schema is valid, raises an error otherwise.
        :raise MigrationMissingError, MigrationTooHighError
        """
        version = self.get_version()
        schema_version = get_schema_version()
        if version is None or version < schema_version:
            raise MigrationMissingError(
                f"The modelchecker requires at least schema version "
                f"{schema_version}. Current version: {version}."
            )

        if version > schema_version:
            warnings.warn(
                f"The database version is higher than the modelchecker "
                f"({version} > {schema_version}). This may lead to unexpected "
                f"results. "
            )
        return True

    def get_missing_tables(self):
        pass

    def get_missing_columns(self):
        pass

    def set_views(self):
        """(Re)create views in the spatialite according to the latest definitions."""
        version = self.get_version()
        schema_version = get_schema_version()
        if version != schema_version:
            raise MigrationMissingError(
                f"Setting views requires schema version "
                f"{schema_version}. Current version: {version}."
            )

        _, file_version = get_spatialite_version(self.db)

        _recreate_views(self.db, file_version)

    def set_spatial_indexes(self):
        """(Re)create spatial indexes in the spatialite according to the latest definitions."""
        version = self.get_version()
        schema_version = get_schema_version()
        if version != schema_version:
            raise MigrationMissingError(
                f"Setting views requires schema version "
                f"{schema_version}. Current version: {version}."
            )

        _ensure_spatial_indexes(self.db)

    def upgrade_spatialite_version(self):
        """Upgrade the version of the spatialite file to the version of the
        current spatialite library.

        Does nothing if the current file version > 3 or if the current library
        version is not 4 or 5.

        Raises UpgradeFailedError if there are any SQL constraints violated.
        """
        lib_version, file_version = get_spatialite_version(self.db)
        if file_version == 3 and lib_version in (4, 5):
            self.validate_schema()

            with self.db.file_transaction(start_empty=True) as work_db:
                _upgrade_database(work_db, revision="head", unsafe=True)
                _recreate_views(work_db, file_version=4)
                copy_models(self.db, work_db, self.declared_models)



"""make settings nullable

Revision ID: 0206
Revises: 0205
Create Date: 2021-12-16 13:30:00

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = "0206"
down_revision = "0205"
branch_labels = None
depends_on = None


MIGRATION_QUERIES = """
UPDATE v2_control_measure_map SET weight = 1.0 WHERE weight IS NULL;
"""


def upgrade():
    for q in MIGRATION_QUERIES.split(";"):
        op.execute(q)

    with op.batch_alter_table("v2_control_measure_map") as batch_op:
        batch_op.alter_column("object_id", nullable=False)
        batch_op.alter_column("object_type", nullable=False)
        batch_op.alter_column("weight", nullable=False)

    with op.batch_alter_table("v2_control_memory") as batch_op:
        batch_op.alter_column("measure_variable", nullable=False)
        batch_op.alter_column("action_type", nullable=False)
        batch_op.alter_column("action_value", nullable=False)
        batch_op.alter_column("target_type", nullable=False)
        batch_op.alter_column("target_id", nullable=False)

    with op.batch_alter_table("v2_control_table") as batch_op:
        batch_op.alter_column("measure_variable", nullable=False)
        batch_op.alter_column("action_type", nullable=False)
        batch_op.alter_column("action_table", nullable=False)
        batch_op.alter_column("target_type", nullable=False)
        batch_op.alter_column("target_id", nullable=False)

    with op.batch_alter_table("v2_control") as batch_op:
        batch_op.alter_column("control_type", nullable=False)


def downgrade():
    pass
